from .myqt import QT
import pyqtgraph as pg

import numpy as np

class MyViewBox(pg.ViewBox):
    pass
    

class ProbeGeometryView(QT.QWidget):
    
    def __init__(self, parent = None, channel_groups=None):
        QT.QWidget.__init__(self, parent)
        
        self.channel_groups = channel_groups
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        
        self.combo_chan_grp = QT.QComboBox()
        h.addWidget(self.combo_chan_grp)
        self.combo_chan_grp.clear()
        self.combo_chan_grp.addItems([str(k) for k in self.channel_groups.keys()])
        self.combo_chan_grp.currentIndexChanged .connect(self.on_chan_grp_change)
        
        self.checkbox = QT.QCheckBox('flip_bottom_up')
        h.addWidget(self.checkbox)
        self.checkbox.stateChanged.connect(self.refresh)
        
        #~ self.combo_chan_grp.blockSignals(True)
        #~ self.combo_chan_grp.blockSignals(False)
        #~ self.on_chan_grp_change()

        
        self.graphicsview = pg.GraphicsView()
        self.layout.addWidget(self.graphicsview)
        

        self.viewBox = MyViewBox()
        self.viewBox.disableAutoRange()
        
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        self.plot.showAxis('left', False)
        
        self.refresh()

    
    def on_chan_grp_change(self):
        self.refresh()
        
    def refresh(self, v=None):
        self.plot.clear()
        
        flip_bottom_up = self.checkbox.checkState()
        
        
        chan_grp = int(self.combo_chan_grp.currentText())
        channel_group = self.channel_groups[chan_grp]
        if channel_group['geometry'] is None:
            return
        
        
        geometry = [ channel_group['geometry'][chan] for chan in channel_group['channels'] ]
        geometry = np.array(geometry, dtype='float64')
        
        if flip_bottom_up:
            geometry[:, 1] *= -1.
        
        for c, chan in enumerate(channel_group['channels']):
            x, y = geometry[c]
            
            #~ name = '{}: {}'.format(c, chan)
            name = '{}'.format(chan)
            itemtxt = pg.TextItem(name, anchor=(.5,.5))
            self.plot.addItem(itemtxt)
            itemtxt.setPos(x, y)

        margin = 100
        
        self.plot.setXRange(np.min(geometry[:, 0])-margin, np.max(geometry[:, 0])+margin)
        self.plot.setYRange(np.min(geometry[:, 1])-margin, np.max(geometry[:, 1])+margin)
            

    #~ for c, chan in enumerate(channels):
        #~ x, y = geometry[c]
        #~ ax.plot([x], [y], marker='o', color='r')
        #~ ax.text(x, y, '{}: {}'.format(c, chan),  size=20)




from .myqt import QT
import pyqtgraph as pg

import numpy as np
import time

from .base import WidgetBase
from .tools import TimeSeeker
from ..tools import median_mad
from ..dataio import _signal_types
from ..peeler_tools import make_prediction_signals


class MyViewBox(pg.ViewBox):
    doubleclicked = QT.pyqtSignal()
    gain_zoom = QT.pyqtSignal(float)
    xsize_zoom = QT.pyqtSignal(float)
    def __init__(self, *args, **kwds):
        pg.ViewBox.__init__(self, *args, **kwds)
        #~ self.disableAutoRange()
    def mouseClickEvent(self, ev):
        ev.accept()
    def mouseDoubleClickEvent(self, ev):
        self.doubleclicked.emit()
        ev.accept()
    def mouseDragEvent(self, ev):
        ev.ignore()
    def wheelEvent(self, ev, axis=None):
        if ev.modifiers() == QT.Qt.ControlModifier:
            z = 10 if ev.delta()>0 else 1/10.
        else:
            z = 1.3 if ev.delta()>0 else 1/1.3
        self.gain_zoom.emit(z)
        ev.accept()
    def mouseDragEvent(self, ev):
        ev.accept()
        self.xsize_zoom.emit((ev.pos()-ev.lastPos()).x())


class BaseTraceViewer(WidgetBase):
    
    _params = [{'name': 'auto_zoom_on_select', 'type': 'bool', 'value': True },
                       {'name': 'zoom_size', 'type': 'float', 'value':  0.08, 'step' : 0.001 },
                      {'name': 'plot_threshold', 'type': 'bool', 'value':  True },
                      {'name': 'alpha', 'type': 'float', 'value' : 0.8, 'limits':(0, 1.), 'step':0.05 },
                      {'name': 'xsize_max', 'type': 'float', 'value': 4.0, 'step': 1.0, 'limits':(1.0, np.inf)},
                      ]
    
    def __init__(self,controller=None, signal_type='initial', parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
    
        self.dataio = controller.dataio
        self.signal_type = signal_type
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        self.create_toolbar()
        
        
        # create graphic view and 2 scroll bar
        g = QT.QGridLayout()
        self.layout.addLayout(g)
        self.scroll_chan = QT.QScrollBar()
        g.addWidget(self.scroll_chan, 0,0)
        self.scroll_chan.valueChanged.connect(self.on_scroll_chan)
        self.graphicsview = pg.GraphicsView()
        g.addWidget(self.graphicsview, 0,1)
        self.initialize_plot()
        self.scroll_time = QT.QScrollBar(orientation=QT.Qt.Horizontal)
        g.addWidget(self.scroll_time, 1,1)
        self.scroll_time.valueChanged.connect(self.on_scroll_time)
        
        #handle time by segments
        self.time_by_seg = np.array([0.]*self.dataio.nb_segment, dtype='float64')
        
        self.change_segment(0)
        self.refresh()
    
    _default_color = QT.QColor( 'white')
    
    def create_toolbar(self):
        tb = self.toolbar = QT.QToolBar()
        
        #Segment selection
        self.combo_seg = QT.QComboBox()
        tb.addWidget(self.combo_seg)
        self.combo_seg.addItems([ 'Segment {}'.format(seg_num) for seg_num in range(self.dataio.nb_segment) ])
        self._seg_pos = 0
        self.seg_num = self._seg_pos
        self.combo_seg.currentIndexChanged.connect(self.on_combo_seg_changed)
        tb.addSeparator()
        
        self.combo_type = QT.QComboBox()
        tb.addWidget(self.combo_type)
        self.combo_type.addItems([ signal_type for signal_type in _signal_types ])
        self.combo_type.setCurrentIndex(_signal_types.index(self.signal_type))
        self.combo_type.currentIndexChanged.connect(self.on_combo_type_changed)

        # time slider
        self.timeseeker = TimeSeeker(show_slider=False)
        tb.addWidget(self.timeseeker)
        self.timeseeker.time_changed.connect(self.seek)
        
        # winsize
        self.xsize = .5
        tb.addWidget(QT.QLabel(u'X size (s)'))
        self.spinbox_xsize = pg.SpinBox(value = self.xsize, bounds = [0.001, self.params['xsize_max']], suffix = 's', siPrefix = True, step = 0.1, dec = True)
        self.spinbox_xsize.sigValueChanged.connect(self.on_xsize_changed)
        tb.addWidget(self.spinbox_xsize)
        tb.addSeparator()
        self.spinbox_xsize.sigValueChanged.connect(self.refresh)
        
        #
        but = QT.QPushButton('auto scale')
        but.clicked.connect(self.auto_scale)
        tb.addWidget(but)
        but = QT.QPushButton('settings')
        but.clicked.connect(self.open_settings)
        tb.addWidget(but)
        self.select_button = QT.QPushButton('select', checkable = True)
        tb.addWidget(self.select_button)
        
        
        self.layout.addWidget(self.toolbar)
        
        self._create_other_toolbar()
    

    def initialize_plot(self):
        self.viewBox = MyViewBox()
        self.plot = pg.PlotItem(viewBox=self.viewBox)
        self.graphicsview.setCentralItem(self.plot)
        self.plot.hideButtons()
        self.plot.showAxis('left', False)
        
        self.viewBox.gain_zoom.connect(self.gain_zoom)
        self.viewBox.xsize_zoom.connect(self.xsize_zoom)
        
        self.visible_channels = np.zeros(self.controller.nb_channel, dtype='bool')
        self.max_channel = min(16, self.controller.nb_channel)
        #~ self.max_channel = min(5, self.controller.nb_channel)
        if self.controller.nb_channel>self.max_channel:
            self.visible_channels[:self.max_channel] = True
            self.scroll_chan.show()
            self.scroll_chan.setMinimum(0)
            self.scroll_chan.setMaximum(self.controller.nb_channel-self.max_channel)
            self.scroll_chan.setPageStep(self.max_channel)
        else:
            self.visible_channels[:] = True
            self.scroll_chan.hide()
            
        self.signals_curve = pg.PlotCurveItem(pen='#7FFF00', connect='finite')
        self.plot.addItem(self.signals_curve)

        self.scatter = pg.ScatterPlotItem(size=10, pxMode = True)
        self.plot.addItem(self.scatter)
        self.scatter.sigClicked.connect(self.scatter_item_clicked)
        
        self.channel_labels = []
        self.threshold_lines =[]
        for i, chan_name in enumerate(self.controller.channel_names):
            #TODO label channels
            label = pg.TextItem('{}: {}'.format(i, chan_name), color='#FFFFFF', anchor=(0, 0.5), border=None, fill=pg.mkColor((128,128,128, 180)))
            self.plot.addItem(label)
            self.channel_labels.append(label)
        
        
        for i in range(self.max_channel):
            tc = pg.InfiniteLine(angle = 0., movable = False, pen = pg.mkPen(color=(128,128,128, 120)))
            tc.setPos(0.)
            self.threshold_lines.append(tc)
            self.plot.addItem(tc)
            tc.hide()
        
        pen = pg.mkPen(color=(128,0,128, 120), width=3, style=QT.Qt.DashLine)
        self.selection_line = pg.InfiniteLine(pos = 0., angle=90, movable=False, pen = pen)
        self.plot.addItem(self.selection_line)
        self.selection_line.hide()
        
        self._initialize_plot()
        
        self.gains = None
        self.offsets = None

    def prev_segment(self):
        self.change_segment(self._seg_pos - 1)
        
    def next_segment(self):
        self.change_segment(self._seg_pos + 1)

    def change_segment(self, seg_pos):
        #TODO: dirty because now seg_pos IS seg_num
        self._seg_pos  =  seg_pos
        if self._seg_pos<0:
            self._seg_pos = self.dataio.nb_segment-1
        if self._seg_pos == self.dataio.nb_segment:
            self._seg_pos = 0
        self.seg_num = self._seg_pos
        self.combo_seg.setCurrentIndex(self._seg_pos)
        
        length = self.dataio.get_segment_length(self.seg_num)
        t_start=0.
        t_stop = length/self.dataio.sample_rate
        self.timeseeker.set_start_stop(t_start, t_stop, seek = False)

        self.scroll_time.setMinimum(0)
        self.scroll_time.setMaximum(length)
        
        if self.isVisible():
            self.refresh()
    
    def on_params_changed(self):
        
        # adjust xsize spinbox bounds, and adjust xsize if out of bounds
        self.spinbox_xsize.opts['bounds'] = [0.001, self.params['xsize_max']]
        if self.xsize > self.params['xsize_max']:
            self.spinbox_xsize.sigValueChanged.disconnect(self.on_xsize_changed)
            self.spinbox_xsize.setValue(self.params['xsize_max'])
            self.xsize = self.params['xsize_max']
            self.spinbox_xsize.sigValueChanged.connect(self.on_xsize_changed)
        
        self.refresh()
    
    def on_combo_seg_changed(self):
        s =  self.combo_seg.currentIndex()
        self.change_segment(s)
    
    def on_combo_type_changed(self):
        s =  self.combo_type.currentIndex()
        self.signal_type = _signal_types[s]
        self.estimate_auto_scale()
        self.change_segment(self._seg_pos)
    

    
    def on_xsize_changed(self):
        self.xsize = self.spinbox_xsize.value()
        if self.isVisible():
            self.refresh()
    
    def refresh(self):
        self.seek(self.time_by_seg[self.seg_num])

    def xsize_zoom(self, xmove):
        factor = xmove/100.
        newsize = self.xsize*(factor+1.)
        limits = self.spinbox_xsize.opts['bounds']
        if newsize>0. and newsize<limits[1]:
            self.spinbox_xsize.setValue(newsize)
    
    def auto_scale(self):
        self.estimate_auto_scale()
        self.refresh()
    
    def estimate_auto_scale(self):

        if self.signal_type=='initial':
            i_stop = min(int(60.*self.dataio.sample_rate), self.dataio.get_segment_shape(self.seg_num, chan_grp=self.controller.chan_grp)[0])
            sigs = self.dataio.get_signals_chunk(seg_num=self.seg_num, chan_grp=self.controller.chan_grp,
                    i_start=0, i_stop=i_stop, signal_type=self.signal_type)
            self.med, self.mad = median_mad(sigs.astype('float32'), axis = 0)

        elif self.signal_type=='processed':
            #in that case it should be already normalize
            self.med = np.zeros(self.controller.nb_channel, dtype='float32')
            self.mad = np.ones(self.controller.nb_channel, dtype='float32')
        
        self.factor = 1.
        self.gain_zoom(15.)
    
    def gain_zoom(self, factor_ratio):
        self.factor *= factor_ratio
        self.gains = np.zeros(self.controller.nb_channel, dtype='float32')
        self.offsets = np.zeros(self.controller.nb_channel, dtype='float32')
        n = np.sum(self.visible_channels)
        self.gains[self.visible_channels] = np.ones(n, dtype=float) * 1./(self.factor*max(self.mad))
        self.offsets[self.visible_channels] = np.arange(n)[::-1] - self.med[self.visible_channels]*self.gains[self.visible_channels]
        self.refresh()

    def on_scroll_time(self, val):
        sr = self.controller.dataio.sample_rate
        self.timeseeker.seek(val/sr)
    
    def on_scroll_chan(self, val):
        self.visible_channels[:] = False
        self.visible_channels[val:val+self.max_channel] = True
        self.gain_zoom(1)
        self.refresh()
    
    def center_scrollbar_on_channel(self, c):
        c = c - self.max_channel//2
        c = min(max(c, 0), self.controller.nb_channel-self.max_channel)
        self.scroll_chan.valueChanged.disconnect(self.on_scroll_chan)
        self.scroll_chan.setValue(c)
        self.scroll_chan.valueChanged.connect(self.on_scroll_chan)
        
        self.visible_channels[:] = False
        self.visible_channels[c:c+self.max_channel] = True
        self.gain_zoom(1)
    
    def scatter_item_clicked(self, plot, points):
        if self.select_button.isChecked()and len(points)==1:
            x = points[0].pos().x()
            self.controller.spike_selection[:] = False
            
            pos_click = int(x*self.dataio.sample_rate )
            mask = self.controller.spikes['segment']==self.seg_num
            ind_nearest = np.argmin(np.abs(self.controller.spikes[mask]['index'] - pos_click))
            
            ind_clicked = np.nonzero(mask)[0][ind_nearest]
            self.controller.spike_selection[ind_clicked] = True
            
            self.spike_selection_changed.emit()
            self.refresh()
    
    def on_spike_selection_changed(self):
        ind_selected, = np.nonzero(self.controller.spike_selection)
        n_selected = ind_selected.size
        if self.params['auto_zoom_on_select'] and n_selected==1:
            ind_selected, = np.nonzero(self.controller.spike_selection)
            ind = ind_selected[0]
            peak_ind = self.controller.spikes[ind]['index']
            seg_num = self.controller.spikes[ind]['segment']
            peak_time = peak_ind/self.dataio.sample_rate
            
            if seg_num != self.seg_num:
                self.combo_seg.setCurrentIndex(seg_num)
            
            self.spinbox_xsize.sigValueChanged.disconnect(self.on_xsize_changed)
            self.spinbox_xsize.setValue(self.params['zoom_size'])
            self.xsize = self.params['zoom_size']
            self.spinbox_xsize.sigValueChanged.connect(self.on_xsize_changed)
            
            label = self.controller.spikes[ind]['cluster_label']
            c = self.controller.get_extremum_channel(label)
            
            if c  is None:
                wf = self.controller.dataio.get_signals_chunk(seg_num=seg_num, chan_grp=self.controller.chan_grp,
                        i_start=peak_ind, i_stop=peak_ind+1,
                        signal_type='processed')
                c = np.argmax(np.abs(wf))
            
            self.center_scrollbar_on_channel(c)
            
            self.seek(peak_time)
            
        else:
            self.refresh()
    
    def seek(self, t):
        #~ tp1 = time.perf_counter()
        
        if self.sender() is not self.timeseeker:
            self.timeseeker.seek(t, emit = False)
        
        self.time_by_seg[self.seg_num] = t
        t1,t2 = t-self.xsize/3. , t+self.xsize*2/3.
        t_start = 0.
        sr = self.dataio.sample_rate

        self.scroll_time.valueChanged.disconnect(self.on_scroll_time)
        self.scroll_time.setValue(int(sr*t))
        self.scroll_time.setPageStep(int(sr*self.xsize))
        self.scroll_time.valueChanged.connect(self.on_scroll_time)
        
        ind1 = max(0, int((t1-t_start)*sr))
        ind2 = int((t2-t_start)*sr)

        sigs_chunk = self.dataio.get_signals_chunk(seg_num=self.seg_num, chan_grp=self.controller.chan_grp,
                i_start=ind1, i_stop=ind2, signal_type=self.signal_type)
        
        if sigs_chunk is None: 
            return
        
        if self.gains is None:
            self.estimate_auto_scale()

        
        nb_visible = np.sum(self.visible_channels)
        
        data_curves = sigs_chunk[:, self.visible_channels].T.copy()
        if data_curves.dtype!='float32':
            data_curves = data_curves.astype('float32')
        
        data_curves *= self.gains[self.visible_channels, None]
        data_curves += self.offsets[self.visible_channels, None]
        #~ data_curves[:,0] = np.nan
        
        connect = np.ones(data_curves.shape, dtype='bool')
        connect[:, -1] = 0
        
        times_chunk = np.arange(sigs_chunk.shape[0], dtype='float64')/self.dataio.sample_rate+max(t1, 0)
        times_chunk_tile = np.tile(times_chunk, nb_visible)
        self.signals_curve.setData(times_chunk_tile, data_curves.flatten(), connect=connect.flatten())
        
        
        #channel labels
        i = 1
        for c in range(self.controller.nb_channel):
            if self.visible_channels[c]:
                self.channel_labels[c].setPos(t1, nb_visible-i)
                self.channel_labels[c].show()
                i +=1
            else:
                self.channel_labels[c].hide()

        n = np.sum(self.visible_channels)
        index_visible, = np.nonzero(self.visible_channels)
        for i, c in enumerate(index_visible):
            if self.params['plot_threshold']:
                threshold = self.controller.get_threshold()
                self.threshold_lines[i].setPos(n-i-1 + self.gains[c]*threshold)
                self.threshold_lines[i].show()
            else:
                self.threshold_lines[i].hide()        
        
        
        # plot peak on signal
        all_spikes = self.controller.spikes
        if len(all_spikes)>0:
            keep = (all_spikes['segment']==self.seg_num) & (all_spikes['index']>=ind1) & (all_spikes['index']<ind2)
            spikes_chunk = np.array(all_spikes[keep], copy=True)
            spikes_chunk['index'] -= ind1
            inwindow_ind = spikes_chunk['index']
            inwindow_label = spikes_chunk['cluster_label']
            inwindow_chan = spikes_chunk['channel']
            if np.any(inwindow_chan==-1):
                inwindow_chan = None
            inwindow_selected = np.array(self.controller.spike_selection[keep])

            self.scatter.clear()
            all_x = []
            all_y = []
            all_brush = []
            for k in self.controller.cluster_labels:

                if not self.controller.cluster_visible[k]: continue
                mask = inwindow_label==k
                if np.sum(mask)==0: continue
                
                color = QT.QColor(self.controller.qcolors.get(k, self._default_color))
                color.setAlpha(int(self.params['alpha']*255))
                
                x = times_chunk[inwindow_ind[mask]]
                sigs_chunk_in = sigs_chunk[inwindow_ind[mask], :]
                chan_inds = None
                if k >=0:
                    c = self.controller.get_extremum_channel(k)
                    if c is not None:
                        chan_inds = np.array([c]*np.sum(mask), dtype='int64')
                if chan_inds is None:
                    if inwindow_chan is None:
                        chan_inds = np.argmax(np.abs(sigs_chunk_in), axis=1)
                    else:
                        chan_inds = inwindow_chan[mask]
                mask_visible = self.visible_channels[chan_inds]
                if np.sum(mask_visible)==0: continue
                
                chan_inds = chan_inds[mask_visible]
                x = x[mask_visible]
                y = sigs_chunk_in[mask_visible, :][np.arange(chan_inds.size), chan_inds]*self.gains[chan_inds]+self.offsets[chan_inds]
                
                all_x.append(x)
                all_y.append(y)
                all_brush.append(np.array([pg.mkBrush(color)]*len(x)))
            #~ print()
            #~ print(all_x)
            #~ print(all_y)
            if len(all_x) > 0:
                all_x = np.concatenate(all_x)
                all_y = np.concatenate(all_y)
                all_brush = np.concatenate(all_brush)
                self.scatter.setData(x=all_x, y=all_y, brush=all_brush)
            
            
            if np.sum(inwindow_selected)==1:
                self.selection_line.setPos(times_chunk[inwindow_ind[inwindow_selected]])
                self.selection_line.show()
            else:
                self.selection_line.hide()            
        else:
            spikes_chunk = None
        
        # plot prediction or residuals ...
        self._plot_specific_items(sigs_chunk, times_chunk, spikes_chunk)
        
        #ranges
        self.plot.setXRange( t1, t2, padding = 0.0)
        self.plot.setYRange(-.5, nb_visible-.5, padding = 0.0)
        
        #TODO : do some thing here
        #~ self.graphicsview.repaint()

        #~ tp2 = time.perf_counter()
        #~ print('seek', tp2-tp1)
        


class CatalogueTraceViewer(BaseTraceViewer):
    """
    **Trace viewer** allow to browser raw signal and preprocess signals.
    
    Note that this viewer do not load the entire signals in memory but load
    chunk on demand from HD, that is why depending on the drive it can be
    quite slow. All zoom and scale factor for signals are computed on CPU and
    not on GPU (it is not vispy!!), so this is not the fastest veiwer but many tips
    help user to navigate very efficiently.
    
    What you can do:
      * On the bottom there is a slider over time
      * On the left there is a slider over channels (if nb_channel>16)
      * If several segments you can switch.
      * You can select manually jump to any time
      * You can zoom the X (time) axis with the spinbox ot by **right click** with mouse.
      * **The mouse wheel make a glocal zoom on signal**
      * You can "select" manually a spike with "select" button and this will be show in **ND Scatter**
      * The threshold is a line for each channel. This is very important to why so peak are not detected.
      * Setting button:
        * "auto_zoom_on_select" : auto zoom when select on ndscatter on peak list
        * "zoom_size" in s
        * disable plot threshold
    
    Important:
      * Ths "preprocessed" signal are normalized (robust Z-score) so that the noise variance is 1.
        So the apparent noise **must be** inbetween  [-3, 3]
    
    """
    def __init__(self, controller=None, signal_type = 'processed', parent=None):
        BaseTraceViewer.__init__(self, controller=controller, signal_type=signal_type, parent=parent)
    
    def _create_other_toolbar(self):
        pass
        
    def _initialize_plot(self):
        pass
    
    def _plot_specific_items(self, sigs_chunk, times_chunk, spikes_chunk):
        pass


class PeelerTraceViewer(BaseTraceViewer):
    def __init__(self, controller=None, signal_type = 'processed', parent=None):
        BaseTraceViewer.__init__(self, controller=controller, signal_type=signal_type, parent=parent)
    
    def _create_other_toolbar(self):
        
        self.toolbar2 = QT.QToolBar()
        self.layout.insertWidget(1, self.toolbar2)
        #~ addToolBarBreak
        
        self.plot_buttons = {}
        for name in ['signals', 'prediction', 'residual']:
            self.plot_buttons[name] = but = QT.QPushButton(name,  checkable = True)
            but.clicked.connect(self.refresh)
            self.toolbar2.addWidget(but)
            
            if name in ['signals', 'prediction']:
                but.setChecked(True)
    
    def _initialize_plot(self):
        self.curve_predictions = pg.PlotCurveItem(pen='#FF00FF', connect='finite')
        self.plot.addItem(self.curve_predictions)
        self.curve_residuals = pg.PlotCurveItem(pen='#FFFF00', connect='finite')
        self.plot.addItem(self.curve_residuals)
   
    def _plot_specific_items(self, sigs_chunk, times_chunk, spikes_chunk):
        if spikes_chunk is None: return
        
        #prediction
        #TODO make prediction only on visible!!!! 
        if self.signal_type == 'processed':
            prediction = make_prediction_signals(spikes_chunk, sigs_chunk.dtype, sigs_chunk.shape, self.controller.catalogue)
            residuals = sigs_chunk - prediction
        
        # plots
        nb_visible = np.sum(self.visible_channels)
        times_chunk_tile = np.tile(times_chunk, nb_visible)
        
        def plot_curves(curve, data):
            data = data[:, self.visible_channels].T.copy()
            data *= self.gains[self.visible_channels, None]
            data += self.offsets[self.visible_channels, None]
            #~ data[:,0] = np.nan
            
            connect = np.ones(data.shape, dtype='bool')
            connect[:, -1] = 0
            
            curve.setData(times_chunk_tile, data.flatten(), connect=connect.flatten())
        
        if self.plot_buttons['prediction'].isChecked() and self.signal_type == 'processed':
            plot_curves(self.curve_predictions, prediction)
        else:
            self.curve_predictions.setData([], [])

        if self.plot_buttons['residual'].isChecked() and self.signal_type == 'processed':
            plot_curves(self.curve_residuals, residuals)
        else:
            self.curve_residuals.setData([], [])
        
        if not self.plot_buttons['signals'].isChecked():
            self.signals_curve.setData([], [])




import numpy as np
import time
import os
import shutil

from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor



from tridesclous.tests.testingtools import setup_catalogue


#~ dataset_name='olfactory_bulb'
dataset_name = 'purkinje'
#~ dataset_name='locust'
#~ dataset_name='striatum_rat'



def setup_module():
    cc, params = setup_catalogue('test_cleancluster', dataset_name=dataset_name)
    
    cc.find_clusters(method=params['cluster_method'], **params['cluster_kargs'])
    
    cc.create_savepoint(name='after_find_clusters')
    

def restore_savepoint(dirname, savepoint=None):
    folder = os.path.join(dirname, 'channel_group_0', 'catalogue_constructor')
    savepoint_folder = os.path.join(dirname, 'channel_group_0', 'catalogue_constructor_SAVEPOINT_' + savepoint)
    
    assert os.path.exists(savepoint_folder)
    
    shutil.rmtree(folder)
    shutil.copytree(savepoint_folder, folder)
    

def test_auto_split():
    dirname = 'test_cleancluster'
    
    restore_savepoint(dirname, savepoint='after_find_clusters')
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    
    cc.find_clusters(method='pruningshears')
    print(cc)
    print(cc.n_jobs)
    
    t1 = time.perf_counter()
    cc.cache_some_waveforms()
    t2 = time.perf_counter()
    print('cache_some_waveforms', t2-t1)
    
    t1 = time.perf_counter()
    cc.auto_split_cluster()
    t2 = time.perf_counter()
    print('auto_split_cluster', t2-t1)
    
    print(cc)
    
    cc.create_savepoint(name='after_auto_split')




def test_trash_not_aligned():
    dirname = 'test_cleancluster'
    
    restore_savepoint(dirname, savepoint='after_auto_split')
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    
    t1 = time.perf_counter()
    cc.trash_not_aligned()
    t2 = time.perf_counter()
    print('trash_not_aligned', t2-t1)
    
    cc.create_savepoint(name='after_trash_not_aligned')


def test_auto_merge():
    dirname = 'test_cleancluster'
    
    restore_savepoint(dirname, savepoint='after_trash_not_aligned')
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)

    t1 = time.perf_counter()
    cc.cache_some_waveforms()
    t2 = time.perf_counter()
    print('cache_some_waveforms', t2-t1)

    t1 = time.perf_counter()
    cc.compute_all_centroid()
    t2 = time.perf_counter()
    print('compute_all_centroid', t2-t1)

    
    t1 = time.perf_counter()
    cc.auto_merge_cluster()
    t2 = time.perf_counter()
    print('auto_merge_cluster', t2-t1)
    
    cc.create_savepoint(name='after_auto_merge_cluster')




def test_trash_low_extremum():
    dirname = 'test_cleancluster'
    
    restore_savepoint(dirname, savepoint='after_auto_merge_cluster')
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    
    print(cc)
    
    t1 = time.perf_counter()
    cc.trash_low_extremum()
    t2 = time.perf_counter()
    print('trash_low_extremum', t2-t1)
    
    cc.create_savepoint(name='after_trash_low_extremum')
    print(cc)


def test_trash_small_cluster():
    dirname = 'test_cleancluster'
    
    restore_savepoint(dirname, savepoint='after_trash_low_extremum')
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    
    t1 = time.perf_counter()
    cc.trash_small_cluster()
    t2 = time.perf_counter()
    print('trash_small_cluster', t2-t1)


    
    
    
if __name__ == '__main__':
    setup_module()
    test_auto_split()
    test_trash_not_aligned()
    test_auto_merge()
    test_trash_low_extremum()
    test_trash_small_cluster()
    
    


"""
Here implementation that tale in account the geometry
of the probe to speed up template matching.

"""

import time
import numpy as np
import joblib
from concurrent.futures import ThreadPoolExecutor
import itertools



from .peeler_engine_base import PeelerEngineGeneric

from .peeler_tools import *
from .peeler_tools import _dtype_spike

import sklearn.metrics.pairwise

from .cltools import HAVE_PYOPENCL, OpenCL_Helper
if HAVE_PYOPENCL:
    import pyopencl
    mf = pyopencl.mem_flags

from .peakdetector import get_peak_detector_class

try:
    import numba
    HAVE_NUMBA = True
    from .numba_tools import numba_explore_best_shift, numba_sparse_scalar_product
except ImportError:
    HAVE_NUMBA = False







class PeelerEngineGeometrical(PeelerEngineGeneric):
    def change_params(self,  **kargs):
        PeelerEngineGeneric.change_params(self, **kargs)

    def initialize(self, **kargs):
        PeelerEngineGeneric.initialize(self, **kargs)
        
        # create peak detector
        p = dict(self.catalogue['peak_detector_params'])

        self.peakdetector_engine = p.pop('engine')
        self.peakdetector_method = p.pop('method')
        
        PeakDetector_class = get_peak_detector_class(self.peakdetector_method, self.peakdetector_engine)
        
        chunksize = self.fifo_size-2*self.n_span # not the real chunksize here
        self.peakdetector = PeakDetector_class(self.sample_rate, self.nb_channel,
                                                        chunksize, self.internal_dtype, self.geometry)
        
        self.peakdetector.change_params(**p)
        
        # some attrs
        self.shifts = np.arange(-self.maximum_jitter_shift, self.maximum_jitter_shift+1)
        self.nb_shift = self.shifts.size
        
        #~ self.channel_distances = sklearn.metrics.pairwise.euclidean_distances(self.geometry).astype('float32')
        #~ self.channels_adjacency = {}
        #~ for c in range(self.nb_channel):
            #~ if self.use_sparse_template:
                #~ nearest, = np.nonzero(self.channel_distances[c, :]<self.adjacency_radius_um)
                #~ self.channels_adjacency[c] = nearest
            #~ else:
                #~ self.channels_adjacency[c] = np.arange(self.nb_channel, dtype='int64')
        
        self.mask_already_tested = np.zeros((self.fifo_size, self.nb_channel), dtype='bool')


    def initialize_before_each_segment(self, **kargs):
        PeelerEngineGeneric.initialize_before_each_segment(self, **kargs)
        self.peakdetector.initialize_stream()


    def detect_local_peaks_before_peeling_loop(self):
        # reset tested mask
        self.mask_already_tested[:] = False
        # and detect peak
        self.re_detect_local_peak()
        
        #~ print('detect_local_peaks_before_peeling_loop', self.pending_peaks.size)

    def re_detect_local_peak(self):
        mask = self.peakdetector.get_mask_peaks_in_chunk(self.fifo_residuals)
        if mask.ndim ==1:
            #~ mask &= ~self.mask_already_tested[self.n_span:-self.n_span, 0]
            sample_indexes,   = np.nonzero(mask)
            sample_indexes += self.n_span
            tested = self.mask_already_tested[sample_indexes, 0]
            sample_indexes = sample_indexes[~tested]
            chan_indexes = np.zeros(sample_indexes.size, dtype='int64')
        else:
            #~ mask &= ~self.mask_already_tested[self.n_span:-self.n_span, :]
            sample_indexes, chan_indexes  = np.nonzero(mask)
            sample_indexes += self.n_span
            tested = self.mask_already_tested[sample_indexes, chan_indexes]
            sample_indexes = sample_indexes[~tested]
            chan_indexes = chan_indexes[~tested]
        
        
        amplitudes = np.abs(self.fifo_residuals[sample_indexes, chan_indexes])
        order = np.argsort(amplitudes)[::-1]
        
        dtype_peak = [('sample_index', 'int32'), ('chan_index', 'int32'), ('peak_value', 'float32')]
        self.pending_peaks = np.zeros(sample_indexes.size, dtype=dtype_peak)
        self.pending_peaks['sample_index'] = sample_indexes
        self.pending_peaks['chan_index'] = chan_indexes
        self.pending_peaks['peak_value'] = amplitudes
        self.pending_peaks = self.pending_peaks[order]
        #~ print('re_detect_local_peak', self.pending_peaks.size)

    def select_next_peak(self):
        #~ print(len(self.pending_peaks))
        if len(self.pending_peaks)>0:
            sample_ind, chan_ind, ampl = self.pending_peaks[0]
            self.pending_peaks = self.pending_peaks[1:]
            return sample_ind, chan_ind
        else:
            return LABEL_NO_MORE_PEAK, None

    def on_accepted_spike(self, sample_ind, cluster_idx, jitter):
        # remove spike prediction from fifo residuals
        #~ t1 = time.perf_counter()
        pos, pred = make_prediction_one_spike(sample_ind, cluster_idx, jitter, self.fifo_residuals.dtype, self.catalogue)
        #~ t2 = time.perf_counter()
        #~ print('    make_prediction_one_spike', (t2-t1)*1000)
        
        #~ t1 = time.perf_counter()
        self.fifo_residuals[pos:pos+self.peak_width_long, :] -= pred
        #~ t2 = time.perf_counter()
        #~ print('     self.fifo_residuals -', (t2-t1)*1000)

        # this prevent search peaks in the zone until next "reset_to_not_tested"
        #~ t1 = time.perf_counter()
        self.clean_pending_peaks_zone(sample_ind, cluster_idx)
        #~ t2 = time.perf_counter()
        #~ print('     self.clean_pending_peaks_zone -', (t2-t1)*1000)


    def clean_pending_peaks_zone(self, sample_ind, cluster_idx):
        # TODO test with sparse_mask_level3s!!!!!
        mask = self.sparse_mask_level1[cluster_idx, :]

        
        #~ t1 = time.perf_counter()
        #~ keep = np.zeros(self.pending_peaks.size, dtype='bool')
        #~ for i, peak in enumerate(self.pending_peaks):
            #~ in_zone = mask[peak['chan_index']] and \
                                #~ (peak['sample_index']+self.n_left)<sample_ind and \
                                #~ sample_ind<(peak['sample_index']+self.n_right)
            #~ keep[i] = not(in_zone)
        
        peaks = self.pending_peaks
        in_zone = mask[peaks['chan_index']] &\
                            ((peaks['sample_index']+self.n_left)<sample_ind) & \
                            ((peaks['sample_index']+self.n_right)>sample_ind)
        keep = ~ in_zone
        #~ t2 = time.perf_counter()
        #~ print('     clean_pending_peaks_zone loop', (t2-t1)*1000)
        
        self.pending_peaks = self.pending_peaks[keep]
        
        #~ print('clean_pending_peaks_zone', self.pending_peaks.size)
    
    def set_already_tested(self, sample_ind, peak_chan):
        self.mask_already_tested[sample_ind, peak_chan] = True

    def reset_to_not_tested(self, good_spikes):
        for spike in good_spikes:
            # each good spike can remove from
            cluster_idx = self.catalogue['label_to_index'][spike.cluster_label]
            chan_mask = self.sparse_mask_level1[cluster_idx, :]
            self.mask_already_tested[spike.index + self.n_left_long:spike.index + self.n_right_long][:, chan_mask] = False
        
        self.re_detect_local_peak()
        

    def get_no_label_peaks(self):
        mask = self.peakdetector.get_mask_peaks_in_chunk(self.fifo_residuals)
        nolabel_indexes, chan_indexes = np.nonzero(mask)
        #~ nolabel_indexes, chan_indexes = np.nonzero(~self.mask_not_already_tested)
        
        nolabel_indexes += self.n_span
        nolabel_indexes = nolabel_indexes[nolabel_indexes<(self.chunksize+self.n_span)]
        bad_spikes = np.zeros(nolabel_indexes.shape[0], dtype=_dtype_spike)
        bad_spikes['index'] = nolabel_indexes
        bad_spikes['cluster_label'] = LABEL_UNCLASSIFIED
        return bad_spikes

    def get_best_template(self, left_ind, chan_ind):

        full_waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:]
        
        centers0 = self.catalogue['centers0']
        projections = self.catalogue['projections']

        strict_low = self.catalogue['boundaries'][:, 0]
        strict_high = self.catalogue['boundaries'][:, 1]
        flexible_low = self.catalogue['boundaries'][:, 2]
        flexible_high = self.catalogue['boundaries'][:, 3]
        
        
        n = centers0.shape[0]
        flat_waveform = full_waveform.flatten()
        flat_centers0 = centers0.reshape(n, -1)
        
        #~ scalar_products = np.zeros(n, dtype='float32')
        #~ for i in range(n):
            #~ sp = np.sum((flat_waveform - flat_centers0[i, :]) * projections[i, :])
            #~ scalar_products[i] = sp
        #~ scalar_products = np.sum((flat_waveform[np.newaxis, :] - flat_centers0[:, :]) * projections[:, :], axis=1)
        #~ print(scalar_products)
        
        #~ t1 = time.perf_counter()
        scalar_products = numba_sparse_scalar_product(self.fifo_residuals, left_ind, centers0, projections, chan_ind,
                    self.sparse_mask_level1, )
        #~ t2 = time.perf_counter()
        #~ print('numba_sparse_scalar_product', (t2-t1)*1000)

        #~ print(scalar_products)
        
        
        possible_idx, = np.nonzero((scalar_products < strict_high) & (scalar_products > strict_low))
        #~ possible_idx, = np.nonzero((scalar_products < flexible_high) & (scalar_products > flexible_low))
        
        #~ print('possible_idx', possible_idx)
        #~ print('scalar_products[possible_idx]', scalar_products[possible_idx])
        
        
        #~ do_plot = False
        if len(possible_idx) == 1:
            extra_idx = None
            candidates_idx =possible_idx
        elif len(possible_idx) == 0:
            #~ extra_idx, = np.nonzero((np.abs(scalar_products) < 0.5))
            extra_idx, = np.nonzero((scalar_products < flexible_high) & (scalar_products > flexible_low))
            #~ if len(extra_idx) ==0:
                # give a try to very far ones.
                #~ extra_idx, = np.nonzero((np.abs(scalar_products) < 1.))
                #~ print('extra_idx', extra_idx)
            #~ if len(extra_idx) ==0:
                #~ candidates_idx = []
            #~ else:
                #~ candidates_idx = extra_idx
            candidates_idx = extra_idx
            #~ candidates_idx =possible_idx
            #~ pass
        elif len(possible_idx) > 1 :
            extra_idx = None
            candidates_idx = possible_idx
        
        debug_plot_change = False
        if len(candidates_idx) > 0:
            #~ t1 = time.perf_counter()
            candidates_idx = np.array(candidates_idx, dtype='int64')
            common_mask = np.sum(self.sparse_mask_level3[candidates_idx, :], axis=0) > 0
            shift_scalar_product, shift_distance = numba_explore_best_shift(self.fifo_residuals, left_ind, self.catalogue['centers0'],
                            self.catalogue['projections'], candidates_idx, self.maximum_jitter_shift, common_mask, self.sparse_mask_level1)
            #~ i0, i1 = np.unravel_index(np.argmin(np.abs(shift_scalar_product), axis=None), shift_scalar_product.shape)
            i0, i1 = np.unravel_index(np.argmin(shift_distance, axis=None), shift_distance.shape)
            #~ best_idx = candidates_idx[i0]
            shift = self.shifts[i1]
            cluster_idx = candidates_idx[i0]
            final_scalar_product = shift_scalar_product[i0, i1]
            #~ t2 = time.perf_counter()
            #~ print('numba_explore_best_shift', (t2-t1)*1000)



            
            
            #~ print('shift', shift)
            #~ print('cluster_idx', cluster_idx)
            #~ print('final_scalar_product', final_scalar_product)
            
            
            if np.abs(shift) == self.maximum_jitter_shift:
                cluster_idx = None
                shift = None
                final_scalar_product = None
                #~ print('maximum_jitter_shift >> cluster_idx = None ')
                #~ do_plot = True
            #~ i0_bis, i1_bis = np.unravel_index(np.argmin(np.abs(shift_scalar_product), axis=None), shift_scalar_product.shape)
            #~ if i0 != i0_bis:
                
                #~ debug_plot_change = True
                #~ print('Warning')
                #~ print(possible_idx)
                #~ print(shift_scalar_product)
                #~ print(shift_distance)
            
            
            
            
            #~ if best_idx != cluster_idx:
                #~ print('*'*50)
                #~ print('best_idx != cluster_idx', best_idx, cluster_idx)
                #~ print('*'*50)
                #~ cluster_idx = best_idx
                #~ debug_plot_change = True
        else:
            cluster_idx = None
            shift = None
            final_scalar_product = None
            
            #~ import matplotlib.pyplot as plt
            #~ fig, ax = plt.subplots()
            #~ ax.plot(self.shifts, shift_scalar_product.T)
            #~ plt.show()
            
        
        #~ print('ici',)
        

        # DEBUG OMP
        #~ from sklearn.linear_model import orthogonal_mp_gram
        #~ from sklearn.linear_model import OrthogonalMatchingPursuit
        #~ n_nonzero_coefs = 2
        #~ omp = OrthogonalMatchingPursuit(n_nonzero_coefs=n_nonzero_coefs)
        #~ X = self.catalogue['centers0'].reshape(self.catalogue['centers0'].shape[0], -1).T
        #~ waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:].flatten()
        #~ y = waveform
        #~ omp.fit(X, y)
        #~ coef = omp.coef_
        #~ idx_r, = coef.nonzero()
        #~ cluster_idx_omp = np.argmin(np.abs(coef - 1))
        
        
        #~ if cluster_idx_omp != cluster_idx and coef[cluster_idx_omp] > 0.5:
        #~ if True:
        if False:
            
        
        
        
        #~ if cluster_idx in (3,6):
        #~ if do_plot:
        #~ if False:
        #~ if final_scalar_product is not None and np.abs(final_scalar_product) > 0.5:
            
        #~ if True:
        #~ if len(possible_idx) != 1:
        #~ if len(possible_idx) > 1:
        #~ if len(candidates_idx) > 1:
        
        #~ if 7 in possible_idx or  cluster_idx == 7:
        #~ if cluster_idx not in possible_idx and len(possible_idx) > 0:
        #~ if debug_plot_change:
            
            import matplotlib.pyplot as plt
            
            print()
            print('best cluster_idx', cluster_idx)
            print('possible_idx', possible_idx)
            print('extra_idx', extra_idx)
            print(scalar_products[possible_idx])
            print(strict_high[possible_idx])
            
            print('cluster_idx_omp', cluster_idx_omp)

            
            fig, ax = plt.subplots()
            ax.plot(coef)
            if cluster_idx is not None:
                ax.axvline(cluster_idx)
            ax.set_title(f'{cluster_idx} omp {cluster_idx_omp}')
            #~ plt.show()

            
            
            
            
            fig, ax = plt.subplots()
            shift2 = 0 if shift is None else shift
            full_waveform2 = self.fifo_residuals[left_ind+shift2:left_ind+shift2+self.peak_width,:]
            
            ax.plot(full_waveform2.T.flatten(), color='k')
            if shift !=0 and shift is not None:
                ax.plot(full_waveform.T.flatten(), color='grey', ls='--')
            
            for idx in candidates_idx:
                ax.plot(self.catalogue['centers0'][idx, :].T.flatten(), color='m')
            
            ax.plot(self.catalogue['centers0'][cluster_idx_omp, :].T.flatten(), color='y')
            
            
            
                
            if cluster_idx is not None:
                ax.plot(self.catalogue['centers0'][cluster_idx, :].T.flatten(), color='c', ls='--')
            ax.set_title(f'best {cluster_idx} shift {shift} possible_idx {possible_idx}')
            
            if shift is not None:
                fig, ax = plt.subplots()
                #~ ax.plot(self.shifts, np.abs(shift_scalar_product).T)
                ax.plot(self.shifts, shift_scalar_product.T)
                ax.axhline(0)
                

                fig, ax = plt.subplots()
                ax.plot(self.shifts, np.abs(shift_distance).T)
            
            plt.show()
        
        
        best_template_info = {'nb_candidate' : len(candidates_idx), 'final_scalar_product':final_scalar_product}
        
        
        return cluster_idx, shift, best_template_info


    def accept_tempate(self, left_ind, cluster_idx, jitter, best_template_info):
        if jitter is None:
            # this must have a jitter
            jitter = 0
        
        #~ if np.abs(jitter) > (self.maximum_jitter_shift - 0.5):
            #~ return False
        
        strict_low = self.catalogue['boundaries'][:, 0]
        strict_high = self.catalogue['boundaries'][:, 1]
        flexible_low = self.catalogue['boundaries'][:, 2]
        flexible_high = self.catalogue['boundaries'][:, 3]
        
        
        #~ flat_waveform = full_waveform.flatten()
        #~ sp2 = np.sum((flat_waveform - centers0[cluster_idx, :].flatten()) * projections[cluster_idx, :])
        sp = best_template_info['final_scalar_product']
        nb_candidate = best_template_info['nb_candidate']
        
        if nb_candidate == 1:
            
            #~ accept_template = strict_low[cluster_idx] < sp < strict_high[cluster_idx]
            accept_template = flexible_low[cluster_idx] < sp < flexible_high[cluster_idx]
            
        else:
            accept_template = flexible_low[cluster_idx] < sp < flexible_high[cluster_idx]
        
                
            # waveform L2 on mask
            #~ full_waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:]
            #~ wf = full_waveform[:, mask]
            
            # prediction with interpolation
            #~ _, pred_wf = make_prediction_one_spike(left_ind - self.n_left, cluster_idx, jitter, self.fifo_residuals.dtype, self.catalogue, long=False)
            #~ pred_wf = pred_wf[:, mask]
        
            #~ dist = (pred_wf - wf) ** 2
            
            
            # criteria per channel
            #~ residual_nrj_by_chan = np.sum(dist, axis=0)
            #~ wf_nrj = np.sum(wf**2, axis=0)
            #~ weight = self.weight_per_template_dict[cluster_idx]
            #~ crietria_weighted = (wf_nrj>residual_nrj_by_chan).astype('float') * weight
            #~ accept_template = np.sum(crietria_weighted) >= 0.7 * np.sum(weight)
            
            # criteria per sample
            #~ dist * np.abs(pred_wf) < 
            #~ dist_w = dist / np.abs(pred_wf)
            #~ gain = (dist < wf**2).astype('float') * np.abs(pred_wf) / np.sum(np.abs(pred_wf))
            #~ gain = (wf / pred_wf - 1) * np.abs(pred_wf) / np.sum(np.abs(pred_wf))
            #~ gain = (pred_wf**2 / wf**1 - 1) * np.abs(pred_wf) / np.sum(np.abs(pred_wf))
            #~ accept_template = np.sum(gain) > 0.8
            #~ accept_template = np.sum(gain) > 0.7
            #~ accept_template0 = np.sum(gain) > 0.6
            #~ accept_template = np.sum(gain) > 0.5
            
            # criteria max residual
            #~ max_res = np.max(np.abs(pred_wf - wf))
            #~ max_pred = np.max(np.abs(pred_wf))
            #~ accept_template1 = max_pred > max_res
            
            
            

            
            
            
            #~ accept_template = False
            
            # debug
            #~ limit_sp =self.catalogue['sp_normed_limit'][cluster_idx, :]
            #~ sp = np.sum(self.catalogue['centers0_normed'] * full_waveform * self.catalogue['template_weight'])
            #~ print('limit_sp', limit_sp, 'sp', sp)
            
            
            
            #~ accept_template = False
            #~ immediate_accept = False
            
            # DEBUG always refuse!!!!!
            #~ accept_template = False
        
        
        #~ label = self.catalogue['cluster_labels'][cluster_idx]
        
        # debug
        #~ if label == 13:
            #~ if accept_template and not immediate_accept:
                #~ accept_template = False
        
        # debug
        #~ if label == 13:
            #~ if not hasattr(self, 'count_accept'):
                #~ self.count_accept = {}
                #~ self.count_accept[label] = {'accept_template':0, 'immediate_accept':0, 'not_accepted':0}
            
            #~ if accept_template:
                #~ self.count_accept[label]['accept_template'] += 1
                #~ if immediate_accept:
                    #~ self.count_accept[label]['immediate_accept'] += 1
            #~ else:
                #~ self.count_accept[label]['not_accepted'] += 1
            #~ print(self.count_accept)
            
        #~ if self._plot_debug:
        #~ if not accept_template and label in []:
        #~ if not accept_template:
        #~ if accept_template:
        #~ if True:
        if False:
            
        #~ if not immediate_accept:
        #~ if immediate_accept:
        #~ if immediate_accept:
        #~ if label == 7 and not accept_template:
        #~ if label == 7:
        #~ if label == 121:
        #~ if label == 5:
        #~ if nb_candidate > 1:
        
        #~ if label == 13 and accept_template and not immediate_accept:
        #~ if label == 13 and not accept_template:
            
        #~ if label in (7,9):
        #~ nears = np.array([ 5813767,  5813767, 11200038, 11322540, 14989650, 14989673, 14989692, 14989710, 15119220, 15830377, 16138346, 16216666, 17078883])
        #~ print(np.abs((left_ind - self.n_left) - nears))
        #~ print(np.abs((left_ind - self.n_left) - nears) < 2)
        #~ if label == 5 and np.any(np.abs((left_ind - self.n_left) - nears) < 50):
            
            #~ if immediate_accept:
            
            import matplotlib.pyplot as plt
            
            mask = self.sparse_mask_level2[cluster_idx]
            full_waveform = self.fifo_residuals[left_ind:left_ind+self.peak_width,:]
            wf = full_waveform[:, mask]
            _, pred_waveform = make_prediction_one_spike(left_ind - self.n_left, cluster_idx, jitter, self.fifo_residuals.dtype, self.catalogue, long=False)
            pred_wf = pred_waveform[:, mask]
    
            if accept_template:
                color = 'g'
            else:
                color = 'r'
            
            #~ if accept_template:
                #~ if immediate_accept:
                    #~ color = 'g'
                #~ else:
                    #~ color = 'c'
            #~ else:
                #~ color = 'r'
            
            #~ if not immediate_accept:
                #~ fig, ax = plt.subplots()
                #~ ax.plot(gain.T.flatten(), color=color)
                #~ ax.set_title('{}'.format(np.sum(gain)))

            #~ fig, ax = plt.subplots()
            #~ ax.plot(feat_centroids.T, alpha=0.5)
            #~ ax.plot(feat_waveform, color='k')

            fig, ax = plt.subplots()
            ax.plot(full_waveform.T.flatten(), color='k')
            ax.plot(pred_waveform.T.flatten(), color=color)
            
            l0, l1 = strict_low[cluster_idx], strict_high[cluster_idx]
            l2, l3 = flexible_low[cluster_idx], flexible_high[cluster_idx]
            title = f'{cluster_idx} {sp:0.3f} lim [{l0:0.3f} {l1:0.3f}] [{l2:0.3f} {l3:0.3f}] {nb_candidate}'
            ax.set_title(title)
                
            #~ fig, ax = plt.subplots()
            #~ ax.plot(wf.T.flatten(), color='k')
            #~ ax.plot(pred_wf.T.flatten(), color=color)
            
            #~ ax.plot( wf.T.flatten() - pred_wf.T.flatten(), color=color, ls='--')
            
            print()
            print('cluster_idx',cluster_idx, 'accept_template', accept_template)
            #~ print(distance, self.distance_limit[cluster_idx])
            #~ print('distance', distance, distance2, 'limit_distance', self.distance_limit[cluster_idx])

            #~ limit_sp =self.catalogue['sp_normed_limit'][cluster_idx, :]
            #~ sp = np.sum(self.catalogue['centers0_normed'] * full_waveform * self.catalogue['template_weight'])
            #~ sp = np.sum(self.catalogue['centers0_normed'] * full_waveform)
            #~ print('limit_sp', limit_sp, 'sp', sp)
            
            #~ if not immediate_accept:
                #~ print('np.sum(gain)', np.sum(gain))


            #~ fig, ax = plt.subplots()
            #~ res = wf - pred_wf
            #~ count, bins = np.histogram(res, bins=150, weights=np.abs(pred_wf))
            #~ ax.plot(bins[:-1], count)
            #~ plt.show()

            
            
            #~ if distance2 >= self.distance_limit[cluster_idx]:
                #~ print(crietria_weighted, weight)
                #~ print(np.sum(crietria_weighted),  np.sum(weight))
            
            #~ ax.plot(full_wf0.T.flatten(), color='y')
            #~ ax.plot( full_wf.T.flatten() - full_wf0.T.flatten(), color='y')
            
            #~ ax.set_title('not accepted')
            plt.show()
        
        return accept_template
    
    
    def _plot_after_inner_peeling_loop(self):
        pass

    def _plot_before_peeling_loop(self):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plot_sigs = self.fifo_residuals.copy()
        self._plot_sigs_before = plot_sigs
        #~ chan_order = np.argsort(self.channel_distances[0, :])
        
        for c in range(self.nb_channel):
        #~ for c in chan_order:
            plot_sigs[:, c] += c*30
        
        ax.plot(plot_sigs, color='k')

        ax.axvline(self.fifo_size - self.n_right_long, color='r')
        ax.axvline(-self.n_left_long, color='r')

        mask = self.peakdetector.get_mask_peaks_in_chunk(self.fifo_residuals)
        sample_inds, chan_inds= np.nonzero(mask)
        sample_inds += self.n_span
        
        ax.scatter(sample_inds, plot_sigs[sample_inds, chan_inds], color='r')
        ax.set_title(f'nb peak {sample_inds.size}')
        
        #~ plt.show()
        
    
    def _plot_label_unclassified(self, left_ind, peak_chan, cluster_idx, jitter):
        return
        import matplotlib.pyplot as plt
        #~ print('LABEL UNCLASSIFIED', left_ind, cluster_idx)
        fig, ax = plt.subplots()
        
        wf = self.fifo_residuals[left_ind:left_ind+self.peak_width, :]
        wf0 = self.catalogue['centers0'][cluster_idx, :, :]
        
        ax.plot(wf.T.flatten(), color='b')
        #~ ax.plot(wf0.T.flatten(), color='g')
        
        ax.set_title(f'label_unclassified {left_ind-self.n_left} {cluster_idx} chan{peak_chan}')
        
        ax.axvline(peak_chan*self.peak_width-self.n_left)
        
        plt.show()

    def _plot_after_peeling_loop(self, good_spikes):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plot_sigs = self.fifo_residuals.copy()
        
        
        for c in range(self.nb_channel):
            plot_sigs[:, c] += c*30
        ax.plot(plot_sigs, color='k')
        
        ax.plot(self._plot_sigs_before, color='b')
        
        ax.axvline(self.fifo_size - self.n_right_long, color='r')
        ax.axvline(-self.n_left_long, color='r')

        mask = self.peakdetector.get_mask_peaks_in_chunk(self.fifo_residuals)
        sample_inds, chan_inds= np.nonzero(mask)
        sample_inds += self.n_span
        ax.scatter(sample_inds, plot_sigs[sample_inds, chan_inds], color='r')
        
        
        
        good_spikes = np.array(good_spikes, dtype=_dtype_spike)
        pred = make_prediction_signals(good_spikes, self.internal_dtype, plot_sigs.shape, self.catalogue, safe=True)
        plot_pred = pred.copy()
        for c in range(self.nb_channel):
            plot_pred[:, c] += c*30
        
        ax.plot(plot_pred, color='m')
        
        plt.show()








import numpy as np


from .tools import median_mad
from .dataio import DataIO
from .catalogueconstructor import CatalogueConstructor
from .tools import make_color_dict, get_neighborhood


try:
    import matplotlib.pyplot as plt
    print('Problem import matplotlib.pyplot as plt')
except:
    pass
    
    
def plot_probe_geometry(dataio, chan_grp=0,  margin=150, channel_number_mode='absolut'):
    channel_group = dataio.channel_groups[chan_grp]
    channels = channel_group['channels']
    geometry = dataio.get_geometry(chan_grp=chan_grp)
    
    fig, ax = plt.subplots()
    for c, chan in enumerate(channels):
        x, y = geometry[c]
        ax.plot([x], [y], marker='o', color='r')
        if channel_number_mode == 'absolut':
            t = '{}'.format(chan)
        elif channel_number_mode == 'local':
            t = '{}'.format(c)
        elif channel_number_mode == 'both':
            t = '{}: {}'.format(c, chan)
        
        ax.text(x, y, t,  size=14, ha='center')

    ax.set_xlim(np.min(geometry[:, 0])-margin, np.max(geometry[:, 0])+margin)
    ax.set_ylim(np.min(geometry[:, 1])-margin, np.max(geometry[:, 1])+margin)
    
    return fig




def plot_signals(dataio_or_cataloguecconstructor, chan_grp=0, seg_num=0, time_slice=(0., 5.), 
            signal_type='initial', with_span=False, with_peaks=False):
    
    if isinstance(dataio_or_cataloguecconstructor, CatalogueConstructor):
        cataloguecconstructor = dataio_or_cataloguecconstructor
        dataio = cataloguecconstructor.dataio
        chan_grp = cataloguecconstructor.chan_grp
    elif isinstance(dataio_or_cataloguecconstructor, DataIO):
        cataloguecconstructor = None
        dataio = dataio_or_cataloguecconstructor
        
    channel_group = dataio.channel_groups[chan_grp]
    channels = channel_group['channels']
    
    i_start = int(time_slice[0]*dataio.sample_rate)
    i_stop = int(time_slice[1]*dataio.sample_rate)
    
    raw_sigs = dataio.get_signals_chunk(seg_num=seg_num, chan_grp=chan_grp,
                i_start=i_start, i_stop=i_stop, signal_type=signal_type)
    
    if signal_type=='initial':
        med, mad = median_mad(raw_sigs)
        sigs = (raw_sigs-med)/mad
        ratioY = 0.3
    elif signal_type=='processed':
        sigs = raw_sigs.copy()
        ratioY = 0.05

    #spread signals
    sigs *= ratioY
    sigs += np.arange(0, len(channels))[np.newaxis, :]
    
    
    
    times = np.arange(sigs.shape[0])/dataio.sample_rate
    fig, ax = plt.subplots()
    ax.plot(times, sigs)
    
    if with_peaks or with_span:
        assert cataloguecconstructor is not None
        peaks = cataloguecconstructor.all_peaks
        
        keep = (peaks['segment']==seg_num) & (peaks['index']>=i_start) & (peaks['index']<i_stop)
        peak_indexes = peaks[keep]['index'].copy()
        peak_indexes -= i_start
        
        if with_peaks:
            for i in range(len(channels)):
                ax.plot(times[peak_indexes], sigs[peak_indexes, i], ls='None', marker='o', color='k')
        
        if with_span:
            d = cataloguecconstructor.info['peak_detector']
            s = d['peak_span_ms'] / 1000.
            for ind in peak_indexes:
                ax.axvspan(times[ind]-s, times[ind]+s, color='b', alpha = .3)
    
    ax.set_yticks([])
    
    return fig


def _prepare_waveform_fig(ax=None, waveforms=None,
                                        flip_bottom_up=False, 
                                        ratioY='auto', deltaX='auto', margin='auto',
                                        **kargs):
    out_kargs = {}
    out_kargs.update(kargs)
    
    out_kargs['waveforms'] = waveforms
    out_kargs['flip_bottom_up'] = flip_bottom_up
    
    
    if ax is None:
        fig, ax = plt.subplots()
    out_kargs['ax'] = ax
    
    if flip_bottom_up:
        geometry = kargs['geometry'].copy()
        geometry[:, 1] *= -1.
        out_kargs['geometry'] = geometry
    #~ print(out_kargs.keys())
    
    geometry = out_kargs['geometry']
    # compute the smallest distance on x and y axis in between electrode
    if np.unique(geometry[:, 0]).size>1:
        xdist = np.min(np.diff(np.sort(np.unique(geometry[:, 0]))))
    else:
        xdist = None
    
    if np.unique(geometry[:, 1]).size>1:
        ydist = np.min(np.diff(np.sort(np.unique(geometry[:, 1]))))
    else:
        ydist = None
    
    if xdist is None:
        if ydist is not None:
            xdist = ydist
        else:
            xdist = 10.
    if ydist is None:
        if xdist is not None:
            ydist = xdist
        else:
            ydist = 10.

    #deltaX is tha width along x axis
    if deltaX =='auto':
        deltaX = xdist/2.5
    
    if ratioY == 'auto':
        m = np.max(np.abs(waveforms))
        ratioY = ydist / m * 0.7
    if margin =='auto':
        margin = 1.5*min(xdist, ydist)
    
    out_kargs['deltaX'] = deltaX
    out_kargs['ratioY'] = ratioY
    out_kargs['margin'] = margin
    
    
    return out_kargs
    


def _plot_wfs(ax=None, waveforms=None, channels=None, geometry=None, 
                        ratioY=None, deltaX=None, linewidth=2., alpha=1.0, color='k', **unused):
        
    wf = waveforms.copy()
    if wf.ndim ==2:
        wf = wf[None, : ,:]
    width = wf.shape[1]

    vect =np.zeros(wf.shape[1]*wf.shape[2])
    wf *= ratioY
    for i, chan in enumerate(channels):
        x, y = geometry[i]
        vect[i*width:(i+1)*width] = np.linspace(x-deltaX, x+deltaX, num=width)
        wf[:, :, i] += y
    
    wf[:, 0,:] = np.nan
    wf = wf.swapaxes(1,2).reshape(wf.shape[0], -1).T
    
    ax.plot(vect, wf, color=color, lw=linewidth, alpha=alpha)


def _enhance_waveform_fig(ax=None, channels=None, geometry=None,
            show_channels=True, channel_number_mode='absolut',
            show_amplitude=False, ratio_mad=5,
            show_ticks=False, flip_bottom_up=True,
            aspect_ratio='equal', xlims=None, ylims=None,
            margin=None, ratioY=None,
            text_size=10, text_color='k',
            **unused):

    if show_channels:
        for c, chan in enumerate(channels):
            x, y = geometry[c, :]
            if channel_number_mode == 'absolut':
                t = '{}'.format(chan)
            elif channel_number_mode == 'local':
                t = '{}'.format(c)
            elif channel_number_mode == 'both':
                t = '{}: {}'.format(c, chan)
            ax.text(x, y, t, color=text_color, size=text_size,ha='center')

    if xlims is None:
        xlim0, xlim1 = np.min(geometry[:, 0])-margin, np.max(geometry[:, 0])+margin
    else:
        xlim0, xlim1 = xlims
    if ylims is None:
        ylim0, ylim1 = np.min(geometry[:, 1])-margin, np.max(geometry[:, 1])+margin
    else:
        ylim0, ylim1 = ylims
    ax.set_xlim(xlim0, xlim1)
    ax.set_ylim(ylim0, ylim1)
    
    if not show_ticks:
        ax.set_xticks([])
        ax.set_yticks([])
    
    if show_amplitude:
        #~ x = xlim0 + margin/10
        x = xlim0 + (xlim1-xlim0)/50
        y = (ylim1+ylim0)/2
        ax.plot([x, x], [y, y+ratioY*ratio_mad], color='k', linewidth=2)
        ax.text(x,y, '{}*MAD'.format(ratio_mad), ha='left')
        

    ax.set_aspect(aspect_ratio)    


def plot_waveforms_with_geometry(waveforms, channels, geometry, **kargs):
    kargs['waveforms'] = waveforms
    kargs['channels'] = channels
    kargs['geometry'] = geometry
    kargs = _prepare_waveform_fig(**kargs)
    _plot_wfs(**kargs)
    _enhance_waveform_fig(**kargs)

    
def plot_waveforms(cataloguecconstructor, labels=None, nb_max=50, alpha=0.4,
                neighborhood_radius=None, **kargs):
    cc = cataloguecconstructor
    channels = np.array(cc.dataio.channel_groups[cc.chan_grp]['channels'])
    geometry = cc.dataio.get_geometry(chan_grp=cc.chan_grp)
    
    kargs['alpha'] = alpha
    
    all_wfs = cc.get_some_waveforms()
    kargs['waveforms'] = all_wfs
    kargs['channels'] = channels
    kargs['geometry'] = geometry
    kargs = _prepare_waveform_fig(**kargs)
    
    if labels is None:
        wfs = all_wfs[:nb_max,:, :]
        kargs['waveforms'] = wfs
        _plot_wfs(**kargs)
        _enhance_waveform_fig(**kargs)
    else:
        if not hasattr(cc, 'colors'):
            cc.refresh_colors()
        
        if isinstance(labels, int):
            labels = [labels]
        
        for label in labels:
            peaks = cc.all_peaks[cc.some_peaks_index]
            keep = peaks['cluster_label'] == label
            wfs = all_wfs[keep][:nb_max]
            
            kargs['color'] = cc.colors.get(label, 'k')
            kargs['waveforms'] = wfs
            _plot_wfs(**kargs)
        
        _enhance_waveform_fig(**kargs)


def plot_centroids(arg0, labels=[], alpha=1, neighborhood_radius=None, **kargs):
    """
    arg0 can be cataloguecconstructor or catalogue (a dict)
    """
    
    if isinstance(labels, int):
        labels = [labels]
    
    if isinstance(arg0, CatalogueConstructor):
        cc = arg0
        dataio = cc.dataio       
        if not hasattr(cc, 'colors'):
            cc.refresh_colors()
        colors = cc.colors
        chan_grp = cc.chan_grp
        
        centroids_wfs = cc.centroids_median
        
        label_inds = []
        for label in labels:
            ind = cc.index_of_label(label)
            label_inds.append(ind)
            
        ratio_mad = cc.info['peak_detector']['relative_threshold']
        
    elif isinstance(arg0, dict) and 'clusters' in arg0:
        catalogue = arg0
        dataio = kargs['dataio']
        chan_grp = catalogue['chan_grp']
        clusters = catalogue['clusters']
        colors = make_color_dict(clusters)
        
        centroids_wfs = catalogue['centers0']
        
        label_inds = []
        for label in labels:
            ind = np.nonzero(clusters['cluster_label']==label)[0][0]
            label_inds.append(ind)
        
        ratio_mad = catalogue['peak_detector_params']['relative_threshold']
        
    else:
        raise(Exception('arg0 must a catalogue constructor or a catalogue dict'))
    
    channels = dataio.channel_groups[chan_grp]['channels']
    geometry =dataio.get_geometry(chan_grp=chan_grp)

    if neighborhood_radius is not None:
        assert len(labels) == 1
        neighborhood = get_neighborhood(geometry, neighborhood_radius)
        extremum_channel = clusters[label_inds[0]]['extremum_channel']
        keep = neighborhood[extremum_channel, :]
        centroids_wfs = centroids_wfs[:,  :, keep]
        channels = np.array(channels)[keep]
        geometry = geometry[keep, :]
    
    kargs['ratio_mad'] = ratio_mad
    kargs['waveforms'] = centroids_wfs[label_inds, :, :]
    kargs['channels'] = channels
    kargs['geometry'] = geometry
    kargs = _prepare_waveform_fig(**kargs)

    for ind, label in zip(label_inds, labels):
        wf = centroids_wfs[ind, :, :]
        kargs['waveforms'] = wf
        kargs['color'] = colors.get(label, 'k')
        _plot_wfs(**kargs)
    
    _enhance_waveform_fig(**kargs)





def plot_waveforms_density(waveforms, bin_min, bin_max, bin_size, ax=None):
    if ax is None:
        fig, ax = plt.subplots()
    
    wf = waveforms
    data = wf.swapaxes(1,2).reshape(wf.shape[0], -1)
    
    bins = np.arange(bin_min, bin_max, bin_size)
    
    hist2d = np.zeros((data.shape[1], bins.size))
    indexes0 = np.arange(data.shape[1])
    
    data_bined = np.floor((data-bin_min)/bin_size).astype('int32')
    data_bined = data_bined.clip(0, bins.size-1)
    
    for d in data_bined:
        hist2d[indexes0, d] += 1
    
    im = ax.imshow(hist2d.T, interpolation='nearest', 
                    origin='lower', aspect='auto', extent=(0, data.shape[1], bin_min, bin_max), cmap='hot')
    
    return im
    
def plot_waveforms_histogram(arg0, label=None, ax=None, channels=None,
            bin_min=None, bin_max=None, bin_size=0.1, units='MAD',
            dataio=None,# usefull when arg0 is catalogue
            ):
    """
    arg0 can be cataloguecconstructor or catalogue (a dict)
    """

    if ax is None:
        fig, ax = plt.subplots()
    
    if isinstance(arg0, CatalogueConstructor):
        cc = arg0
        dataio = cc.dataio
        chan_grp = cc.chan_grp
        
        # take waveforms
        #~ ind = cc.index_of_label(label)
        spike_labels = cc.all_peaks['cluster_label'][cc.some_peaks_index]
        #~ wf = cc.some_waveforms[spike_labels==label]
        ind, = np.nonzero(cc.all_peaks['cluster_label'] == label)
        wf = cc.get_some_waveforms(peaks_index=ind)
        wf = wf[:, :, channels]
        if units in ('uV', 'μV'):
            wf = wf * cc.signals_mads[channels][None, None, :] * dataio.datasource.bit_to_microVolt

        n_left = cc.info['extract_waveforms']['n_left']
        n_right = cc.info['extract_waveforms']['n_right']
    
    elif isinstance(arg0, dict) and 'clusters' in arg0:
        catalogue = arg0
        chan_grp = catalogue['chan_grp']
        clusters = catalogue['clusters']

        n_left = catalogue['n_left']
        n_right = catalogue['n_right']
        
        all_wf = []
        for seg_num in range(dataio.nb_segment):
            # TODO loop over segments
            spikes = dataio.get_spikes(seg_num=seg_num, chan_grp=chan_grp,)
        
            # take waveforms
            #~ spike_labels = spikes['cluster_label']
            spikes = spikes[spikes['cluster_label'] == label]
            sample_indexes = spikes['index']
            
            if sample_indexes.size>1000:
                # limit to 1000 spike by segment
                sample_indexes = np.random.choice(sample_indexes, size=1000)
            sample_indexes = sample_indexes[(sample_indexes>-n_left)]
            
            wf_ = dataio.get_some_waveforms(seg_num=seg_num, chan_grp=chan_grp,
                        peak_sample_indexes=sample_indexes, n_left=n_left, n_right=n_right)
            wf_ = wf_[:, :, channels]
            all_wf.append(wf_)
        wf = np.concatenate(all_wf, axis=0)
        
        if units in ('uV', 'μV'):
            wf = wf * catalogue['signals_mads'][channels][None, None, :] * dataio.datasource.bit_to_microVolt
        
        
        
    else:
        raise(Exception('arg0 must a catalogue constructor or a catalogue dict'))


    if bin_min is None:
        bin_min = np.min(wf) - 1.
    if bin_max is None:
        bin_max = np.max(wf) +1
    if bin_size is None:
        if units=='MAD':
            bin_size = 0.1
        elif units in ('uV', 'μV'):
            bin_size = (bin_max - bin_min) / 500.
    
    #~ data = wf.swapaxes(1,2).reshape(wf.shape[0], -1)
    
    #~ bins = np.arange(bin_min, bin_max, bin_size)
    
    #~ hist2d = np.zeros((data.shape[1], bins.size))
    #~ indexes0 = np.arange(data.shape[1])
    
    #~ data_bined = np.floor((data-bin_min)/bin_size).astype('int32')
    #~ data_bined = data_bined.clip(0, bins.size-1)
    
    #~ for d in data_bined:
        #~ hist2d[indexes0, d] += 1
    
    #~ im = ax.imshow(hist2d.T, interpolation='nearest', 
                    #~ origin='lower', aspect='auto', extent=(0, data.shape[1], bin_min, bin_max), cmap='hot')
                    
    im = plot_waveforms_density(wf, bin_min, bin_max, bin_size, ax=ax)

    peak_width = n_right - n_left
    for c, chan in enumerate(channels):
        abs_chan = dataio.channel_groups[chan_grp]['channels'][chan]
        ax.text(c*peak_width-n_left, 0, '{}'.format(abs_chan),  size=10, ha='center', color='w')
        if c>0:
            ax.axvline((c) * peak_width, color='w')
    
    ax.set_xticks([])

def plot_features_scatter_2d(cataloguecconstructor, labels=None, nb_max=500):
    cc = cataloguecconstructor
    
    all_feat = cc.some_features
    n = all_feat.shape[1]
    
    fig, axs = plt.subplots(nrows=n, ncols=n, sharex=True, sharey=True)
    
    l = []
    if labels is None:
        l.append( (cc.some_features[:nb_max], 'k') )
    else:
        if not hasattr(cc, 'colors'):
            cc.refresh_colors()
        
        if isinstance(labels, int):
            labels = [labels]
        for label in labels:
            peaks = cc.all_peaks[cc.some_peaks_index]
            keep = peaks['cluster_label'] == label
            feat = cc.some_features[keep][:nb_max]
            color = cc.colors.get(label, 'k')
            l.append((feat, color))
    
    for c in range(n):
        for r in range(n):
            ax = axs[r, c]
            
            if c==r:
                for feat, color in l:
                    y, x = np.histogram(feat[:, r], bins=100)
                    ax.plot(x[:-1], y, color=color)
            elif c<r:
                for feat, color in l:
                    ax.plot(feat[:, c], feat[:, r], color=color, markersize=2, ls='None', marker='o')
            else:
                fig.delaxes(ax)



def plot_isi(dataio, catalogue=None, chan_grp=None, label=None, ax=None, bin_min=0, bin_max=100, bin_size=1.):
    """
    bin are in ms
    """
    if ax is None:
        fig, ax = plt.subplots()

    if catalogue is None:
        catalogue = dataio.load_catalogue(chan_grp=chan_grp)
    
    sr = dataio.sample_rate
    
    bins = np.arange(bin_min, bin_max, bin_size)
    
    count = None
    for seg_num in range(dataio.nb_segment):
        spikes = dataio.get_spikes(seg_num=seg_num, chan_grp=chan_grp,)
        spikes = spikes[spikes['cluster_label'] == label]
        sample_indexes = spikes['index']
        
        isi = np.diff(sample_indexes)/ (sr/1000.)
        
        count_, bins = np.histogram(isi, bins=bins)
        if count is None:
            count = count_
        else:
            count += count_
    
    ax.plot(bins[:-1], count, color='k') # TODO color
    


    



"""

.. automodule:: tridesclous.dataio

.. automodule:: tridesclous.catalogueconstructor

.. automodule:: tridesclous.peeler




"""
from .version import version as __version__
import os
import warnings

from .datasets import download_dataset, get_dataset

#dynamic import 
from .datasource import data_source_classes
for c in data_source_classes.values():
    globals()[c.__name__] = c

from .tools import open_prb
from .dataio import DataIO
from .signalpreprocessor import offline_signal_preprocessor, estimate_medians_mads_after_preprocesing
from .catalogueconstructor import CatalogueConstructor
from .cataloguetools import apply_all_catalogue_steps
from .peeler import Peeler
from .cltools import get_cl_device_list, set_default_cl_device

from .autoparams import get_auto_params_for_catalogue, get_auto_params_for_peelers


from .importers import import_from_spykingcircus, import_from_spike_interface

# exclude because import matplotlib

# from .matplotlibplot import *
# from .report import *

# from .gui import *




"""
Here implementation for testing new ideas of peeler.

"""

import time
import numpy as np
import joblib
from concurrent.futures import ThreadPoolExecutor



from .peeler_engine_classic import PeelerEngineClassic

from .peeler_tools import *
from .peeler_tools import _dtype_spike

from .cltools import HAVE_PYOPENCL, OpenCL_Helper
if HAVE_PYOPENCL:
    import pyopencl
    mf = pyopencl.mem_flags


try:
    import numba
    HAVE_NUMBA = True
    from numba import jit


    @jit(parallel=True)
    def numba_loop_sparse_weigthed_dist(waveform, centers,  mask):
        nb_clus, width, nb_chan = centers.shape
        
        rms_waveform_channel = np.sum(waveform**2, axis=0)#.astype('float32')
        waveform_distance = np.zeros((nb_clus,), dtype=np.float32)
        
        for clus in range(nb_clus):
            sum = 0
            for c in range(nb_chan):
                if mask[clus, c]:
                    for s in range(width):
                        d = waveform[s, c] - centers[clus, s, c]
                        sum += d*d
                else:
                    sum +=rms_waveform_channel[c]
            waveform_distance[clus] = sum
        
        return waveform_distance
except ImportError:
    HAVE_NUMBA = False








class PeelerEngineTesting(PeelerEngineClassic):

    #~ def estimate_jitter(self, left_ind, cluster_idx):
        #~ return 0
   
    def accept_tempate(self, left_ind, cluster_idx, jitter):
        
        #~ print('jitter', jitter)
        if np.abs(jitter) > (self.maximum_jitter_shift - 0.5):
            
            #~ print('too big jitter', jitter, np.abs(jitter))
            return False
        
        mask = self.sparse_mask[cluster_idx, :]
        
        shift = -int(np.round(jitter))
        jitter = jitter + shift
        left_ind = left_ind + shift
        if left_ind<0:
            return False
        new_left, pred_wf = make_prediction_one_spike(left_ind - self.n_left, cluster_idx, jitter, self.fifo_residuals.dtype, self.catalogue)
        pred_wf = pred_wf[:, :][:, mask]
        

        
        full_wf0 = self.catalogue['centers0'][cluster_idx,: , :][:, mask]
        #~ full_wf1 = self.catalogue['centers1'][cluster_idx,: , :][:, mask]
        #~ full_wf2 = self.catalogue['centers2'][cluster_idx,: , :][:, mask]
        #~ pred_wf = (full_wf0+jitter*full_wf1+jitter**2/2*full_wf2)
        #~ new_left = left_ind


        #debug
        #~ waveform_no_shift = self.fifo_residuals[left_ind - shift:left_ind- shift +self.peak_width,:]
        #~ full_wf_no_shift = waveform_no_shift[:, :][:, mask]
        

        # waveform L2 on mask
        waveform = self.fifo_residuals[new_left:new_left+self.peak_width,:]
        full_wf = waveform[:, :][:, mask]
        wf_nrj = np.sum(full_wf**2, axis=0)

        #~ thresh_ratio = 0.6
        thresh_ratio = 0.7
        #~ thresh_ratio = 0.8
        
        # criteria per channel
        #~ residual_nrj = np.sum((full_wf-pred_wf)**2, axis=0)
        #~ label = self.catalogue['cluster_labels'][cluster_idx]
        #~ weight = self.weight_per_template[label]
        #~ crietria_weighted = (wf_nrj>residual_nrj).astype('float') * weight
        #~ accept_template = np.sum(crietria_weighted) >= 0.9 * np.sum(weight)
        
        weigth = pred_wf ** 2 # TODO precompute this !!!!!!
        residual = (full_wf-pred_wf)
        s = np.sum((full_wf**2>=residual**2).astype(float) * weigth)
        #~ s = np.sum((pred_wf**2*weigth)>(residual*weigth))
        accept_template = s >np.sum(weigth) * thresh_ratio
        #~ print(s, np.sum(weigth) , np.sum(weigth)  * thresh_ratio)
        
        
        #DEBUG
        label = self.catalogue['cluster_labels'][cluster_idx]
        #~ if label in (0, ):
        if False:
        #~ if True:
            
            #~ print('accept_tempate',accept_template, 'label', label)
            #~ print(wf_nrj>residual_nrj)
            #~ print(weight)
            #~ print(crietria_weighted)
            #~ print(np.sum(crietria_weighted), np.sum(weight), np.sum(crietria_weighted)/np.sum(weight))
            #~ print()
            
            #~ if not accept_template:
                #~ print(wf_nrj>residual_nrj)
                #~ print(weight)
                #~ print(crietria_weighted)
                #~ print()
            print(s, np.sum(weigth) , np.sum(weigth)  * thresh_ratio)
            import matplotlib.pyplot as plt
            fig, axs = plt.subplots(nrows=3, sharex=True)
            axs[0].plot(full_wf.T.flatten(), color='b')
            #~ axs[0].plot(full_wf_no_shift.T.flatten(), color='c')
            
            
            if accept_template:
                axs[0].plot(pred_wf.T.flatten(), color='g')
            else:
                axs[0].plot(pred_wf.T.flatten(), color='r')
            
            #~ axs[0].plot(full_wf0.T.flatten(), color='k')
            
            
            
            axs[0].plot((full_wf-pred_wf).T.flatten(), color='m', lw=1)
            
            axs[1].plot((full_wf**2).T.flatten(), color='b')
            axs[1].plot((residual**2).T.flatten(), color='m')
            
            criterium = (full_wf**2>residual**2).astype(float) * weigth
            axs[2].plot(criterium.T.flatten(), color='k')
            
            plt.show()
            
        
        #~ #ENDDEBUG
        
        
        return accept_template




from tridesclous import get_dataset
from tridesclous.signalpreprocessor import signalpreprocessor_engines, offline_signal_preprocessor

from tridesclous.signalpreprocessor import HAVE_PYOPENCL

import time

import scipy.signal
import numpy as np

from matplotlib import pyplot




def run_online(engine, sigs, sample_rate, chunksize, **params):
    nb_channel = sigs.shape[1]
    
    # precompute medians and mads
    params2 = dict(params)
    params2['normalize'] = False
    sigs_for_noise = offline_signal_preprocessor(sigs, sample_rate, **params2)
    medians = np.median(sigs_for_noise, axis=0)
    mads = np.median(np.abs(sigs_for_noise-medians),axis=0)*1.4826
    params['signals_medians'] = medians
    params['signals_mads'] = mads

    SignalPreprocessorClass = signalpreprocessor_engines[engine]
    signalpreprocessor = SignalPreprocessorClass(sample_rate, nb_channel, chunksize,'float32')
    signalpreprocessor.change_params(**params)
    
    nloop = sigs.shape[0]//chunksize
    
    all_online_sigs = []
    t1 = time.perf_counter()
    for i in range(nloop):
        #~ print(i)
        pos = (i+1)*chunksize
        chunk = sigs[pos-chunksize:pos,:]
        pos2, preprocessed_chunk = signalpreprocessor.process_buffer_stream(pos, chunk)
        if preprocessed_chunk is not None:
            #~ print(preprocessed_chunk)
            all_online_sigs.append(preprocessed_chunk)
    t2 = time.perf_counter()
    print(engine, ' online process time', t2-t1)
    final_sig = np.concatenate(all_online_sigs)
    
    return final_sig


def run_by_buffer(engine, sigs, sample_rate, chunksize, **params):
    nb_channel = sigs.shape[1]
    pad_width = params['pad_width']
    
    # precompute medians and mads
    params2 = dict(params)
    params2['normalize'] = False
    sigs_for_noise = offline_signal_preprocessor(sigs, sample_rate, **params2)
    medians = np.median(sigs_for_noise, axis=0)
    mads = np.median(np.abs(sigs_for_noise-medians),axis=0)*1.4826
    params['signals_medians'] = medians
    params['signals_mads'] = mads

    SignalPreprocessorClass = signalpreprocessor_engines[engine]
    signalpreprocessor = SignalPreprocessorClass(sample_rate, nb_channel, chunksize, 'float32')
    signalpreprocessor.change_params(**params)
    
    nloop = sigs.shape[0]//chunksize
    
    all_chunk_sigs = []
    t1 = time.perf_counter()
    for i in range(nloop):
        #~ print(i)
        i0, i1 = i*chunksize, (i+1)*chunksize
        if i==0:
            chunk_2pad = np.zeros((chunksize+2*pad_width, nb_channel), dtype='float32')
            chunk_2pad[pad_width:, :] = sigs[i0:i1+pad_width,:]
        elif i==(nloop-1):
            chunk_2pad = np.zeros((chunksize+2*pad_width, nb_channel), dtype='float32')
            chunk_2pad[:-pad_width, :] = sigs[i0-pad_width:i1,:]
        else:
            chunk_2pad = sigs[i0-pad_width:i1+pad_width,:].astype('float32')
        
        preprocessed_chunk = signalpreprocessor.process_buffer(chunk_2pad)
        #~ print(preprocessed_chunk)
        
        all_chunk_sigs.append(preprocessed_chunk[pad_width:-pad_width, :])
    
    t2 = time.perf_counter()
    print(engine, 'by buffer process time', t2-t1)
    final_sig = np.concatenate(all_chunk_sigs)
    
    return final_sig
    

    
    




def test_compare_offline_online_engines():
    #~ HAVE_PYOPENCL = True
    if HAVE_PYOPENCL:
        engines = ['numpy', 'opencl']
        #~ engines = [ 'opencl']
        #~ engines = ['numpy']
    else:
        engines = ['numpy']


    # get sigs
    sigs, sample_rate = get_dataset(name='olfactory_bulb')
    print(sample_rate)
    sigs = np.tile(sigs, (1, 20)) #for testing large channels num
    
    nb_channel = sigs.shape[1]
    print('nb_channel', nb_channel)
    
    #params
    chunksize = 1024
    nloop = sigs.shape[0]//chunksize
    sigs = sigs[:chunksize*nloop]
    
    print('sig duration', sigs.shape[0]/sample_rate)
    
    highpass_freq = 500.
    #~ highpass_freq = 300.
    #~ highpass_freq = 100.
    #~ highpass_freq = None
    lowpass_freq = 4000.
    #~ lowpass_freq = None
    #~ smooth_kernel = True
    smooth_size = 0
    
    
    #~ pad_width = int(sample_rate/highpass_freq)*4
    #~ print('pad_width', pad_width)
    
    params = {
                'common_ref_removal' : False,
                'highpass_freq': highpass_freq,
                'lowpass_freq': lowpass_freq,
                'smooth_size':smooth_size,
                'output_dtype': 'float32',
                'normalize' : True,
                'pad_width': 150
                }
    
    t1 = time.perf_counter()
    offline_sig = offline_signal_preprocessor(sigs, sample_rate, **params)
    
    t2 = time.perf_counter()
    print('offline', 'process time', t2-t1)
    
    online_sigs = {}
    by_buffer_sigs = {}
    for engine in engines:
        if engine == 'opencl':
            params['cl_platform_index'] = 0
            params['cl_device_index'] = 0
        online_sigs[engine] = run_online(engine, sigs, sample_rate, chunksize, **params)
        by_buffer_sigs[engine] = run_by_buffer(engine, sigs, sample_rate, chunksize, **params)
        
    
    # remove border for comparison
    min_size = min([online_sigs[engine].shape[0] for engine in engines]) 
    offline_sig = offline_sig[chunksize:min_size]
    for engine in engines:
        online_sigs[engine] = online_sigs[engine][chunksize:min_size]
        
        by_buffer_sigs[engine] = by_buffer_sigs[engine][chunksize:min_size]

    # compare
    for engine in engines:
        print()
        print(engine)
        online_sig = online_sigs[engine]
        residual = np.abs(online_sig.astype('float64')-offline_sig.astype('float64'))
        print('max residual online', np.max(residual))
        print('/', np.mean(np.abs(offline_sig.astype('float64'))))
        residual_normed = residual/np.mean(np.abs(offline_sig.astype('float64')))
        print('max residual_normed', np.max(residual_normed))


        online_sig = by_buffer_sigs[engine]
        residual = np.abs(online_sig.astype('float64')-offline_sig.astype('float64'))
        print('max residual by buffer', np.max(residual))
        print('/', np.mean(np.abs(offline_sig.astype('float64'))))
        residual_normed = residual/np.mean(np.abs(offline_sig.astype('float64')))
        print('max residual_normed', np.max(residual_normed))




        # plot
        #~ # fig, axs = pyplot.subplots(nrows=nb_channel, sharex=True)
        #~ fig, axs = pyplot.subplots(nrows=4, sharex=True)
        # for i in range(nb_channel):
        #~ for i in range(4):
            #~ ax = axs[i]
            #~ ax.plot(residual_normed[:, i], color = 'k')
            #~ ax.plot(residual[:, i], color = 'k')
            #~ ax.plot(offline_sig[:, i], color = 'g')
            #~ ax.plot(online_sig[:, i], color = 'r', ls='--')
            #~ for i in range(nloop):
                #~ ax.axvline(i*chunksize, color='k', alpha=0.4)
        #~ pyplot.show()

        print(np.max(residual_normed))
        #~ print(np.mean(np.abs(offline_sig.astype('float64'))))
        assert np.max(residual_normed)<0.05, 'online differt from offline more than 5%'


def explore_pad_width():

    sigs, sample_rate = get_dataset(name='olfactory_bulb')
    chunksize = 1024
    nloop = sigs.shape[0]//chunksize
    sigs = sigs[:chunksize*nloop]
    
    params = {
                'common_ref_removal' : False,
                'highpass_freq': 300.,
                'lowpass_freq': 4000.,
                'smooth_size':0,
                'output_dtype': 'float32',
                'normalize' : True,
                #~ 'pad_width': 150
                }
    
    offline_sig = offline_signal_preprocessor(sigs, sample_rate, **params)
    
    pad_widths = [ int(sample_rate/params['highpass_freq'])*i for i in range(1, 5)]
    
    online_sigs = {}
    for pad_width in pad_widths:
        print('pad_width', pad_width)
        params['pad_width'] = pad_width
        online_sigs[pad_width] = run_online('numpy', sigs, sample_rate, chunksize, **params)


    # remove border for comparison
    min_size = min([online_sigs[pad_width].shape[0] for pad_width in pad_widths]) 
    offline_sig = offline_sig[chunksize:min_size]
    for pad_width in pad_widths:
        online_sig = online_sigs[pad_width]
        online_sigs[pad_width] = online_sig[chunksize:min_size]


    for pad_width in pad_widths:
        print('pad_width', pad_width)
        
        online_sig = online_sigs[pad_width]
        residual = np.abs(online_sig.astype('float64')-offline_sig.astype('float64'))
        
        residual_normed = residual/np.mean(np.abs(offline_sig.astype('float64')))
        print('   max residual_normed', np.max(residual_normed))



        
    

def test_auto_pad_width():
    sigs, sample_rate = get_dataset(name='olfactory_bulb')

    sigs, sample_rate = get_dataset(name='olfactory_bulb')
    chunksize = 1024
    nloop = sigs.shape[0]//chunksize
    sigs = sigs[:chunksize*nloop]
    
    params = {
                'common_ref_removal' : False,
                'highpass_freq': 300.,
                'lowpass_freq': 4000.,
                'smooth_size':0,
                'output_dtype': 'float32',
                'normalize' : True,
                'pad_width': None,
                }
    
    offline_sig = offline_signal_preprocessor(sigs, sample_rate, **params)
    
    online_sig = run_online('numpy', sigs, sample_rate, chunksize, **params)
    min_size = online_sig.shape[0]
    offline_sig = offline_sig[chunksize:min_size]
    online_sig = online_sig[chunksize:min_size]
    
    residual = np.abs(online_sig.astype('float64')-offline_sig.astype('float64'))
    
    residual_normed = residual/np.mean(np.abs(offline_sig.astype('float64')))
    print('   max residual_normed', np.max(residual_normed))
    
    
    
    # remove border for comparison


def test_smooth_with_filtfilt():
    sigs = np.zeros((100, 1), 'float32')
    #~ sigs[49] = 6
    sigs[50] = 1
    #~ sigs[51] = 5
    #~ sigs += np.random.randn(*sigs.shape)
    
    # smooth with box kernel
    box_size = 3
    kernel = np.ones(box_size)/box_size
    kernel = kernel[:, None]
    sigs_smooth =  scipy.signal.fftconvolve(sigs,kernel,'same')

    # smooth with filter
    #~ coeff = np.array([1/3, 1/3, 1/3, 1,0,0], dtype='float32')
    #~ coeff = np.array([0.5, 0.25, 0.25, 1,0,0], dtype='float32')
    #~ coeff = np.array([0.8, 0.1, 0.1, 1,0,0], dtype='float32')
    b0 = (1./3.)**.5
    b1 = (1-b0)
    b2 =0.
    #~ b2 = (1-b0)/2.
    #~ b0 = .4
    #~ b1 = 0.6
    #~ b2 = 0.
    coeff = np.array([[b0, b1, b2, 1,0,0]], dtype='float32')
    print(coeff)
    coeff = np.tile(coeff, (5, 1))
    sigs_smooth2 = scipy.signal.sosfiltfilt(coeff, sigs, axis=0)
    
    # smooth with LP filter
    coeff = scipy.signal.iirfilter(5, 0.3, analog=False,
                                    btype = 'lowpass', ftype = 'butter', output = 'sos')
    sigs_smooth3 = scipy.signal.sosfiltfilt(coeff, sigs, axis=0)
    
    
    if __name__ == '__main__':
        print(sigs_smooth2.sum())
        print(sigs_smooth2[40:60])
        fig, ax = pyplot.subplots()
        ax.plot(sigs, color='b', label='sig')
        ax.plot(sigs_smooth, color='g', label='box smooth')
        ax.plot(sigs_smooth2, color='r', label='filfilt smooth')
        ax.plot(sigs_smooth3, color='m', label='lp filter')
        
        ax.legend()
        
        pyplot.show()

    
    


    
if __name__ == '__main__':
    test_compare_offline_online_engines()
    #~ test_smooth_with_filtfilt()
    
    #~ explore_pad_width()
    
    #~ test_auto_pad_width()




import numpy as np
import time
import os
import shutil

from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor



from matplotlib import pyplot

from tridesclous.tests.testingtools import setup_catalogue


#~ dataset_name='olfactory_bulb'
#~ dataset_name = 'purkinje'
#~ dataset_name='locust'
dataset_name='striatum_rat'



def setup_module():
    setup_catalogue('test_cluster', dataset_name=dataset_name)

def teardown_module():
    if not(os.environ.get('APPVEYOR') in ('true', 'True')):
        # this fix appveyor teardown_module bug
        shutil.rmtree('test_cluster')


def test_sawchaincut():
    dirname = 'test_cluster'
    
    dataio = DataIO(dirname=dirname)
    cc = CatalogueConstructor(dataio=dataio)
    #~ print(dataio)
    #~ print(cc)
    
    t0 = time.perf_counter()
    cc.find_clusters(method='sawchaincut', print_debug=True)
    t1 = time.perf_counter()
    print('cluster', t1-t0)
    
    #~ print(cc)


    if __name__ == '__main__':
        from tridesclous.gui import mkQApp, CatalogueWindow
        app = mkQApp()
        win = CatalogueWindow(cc)
        win.show()
        app.exec_()


def test_pruningshears():

    dirname = 'test_cluster'
    
    
    dataio = DataIO(dirname=dirname)
    print(dataio)
    cc = CatalogueConstructor(dataio=dataio)
    #~ print(cc.mode)
    
    #~ cc.extract_some_features(method='pca_by_channel')
    #~ print(dataio)
    #~ print(cc)
    
    if dataset_name == 'olfactory_bulb':
        kargs = dict(adjacency_radius_um = 420)
    else:
        kargs = {}
    
    t0 = time.perf_counter()
    #~ cc.find_clusters(method='pruningshears', print_debug=True)
    #~ cc.find_clusters(method='pruningshears', print_debug=True, debug_plot=True, **kargs)
    cc.find_clusters(method='pruningshears', print_debug=False, debug_plot=False, **kargs)
    t1 = time.perf_counter()
    print('cluster', t1-t0)

    if __name__ == '__main__':
        from tridesclous.gui import mkQApp, CatalogueWindow
        app = mkQApp()
        win = CatalogueWindow(cc)
        win.show()
        app.exec_()
    


if __name__ == '__main__':
    setup_module()
    test_sawchaincut()
    test_pruningshears()
    
    



import pytest

from tridesclous import *

import numpy as np
import scipy.signal
import time
import os
import shutil

from tridesclous.dataio import DataIO
from tridesclous.catalogueconstructor import CatalogueConstructor
from tridesclous import Peeler
from tridesclous.peeler_cl import Peeler_OpenCl

from tridesclous.peakdetector import  detect_peaks_in_chunk


from tridesclous.tests.testingtools import setup_catalogue
from tridesclous.tests.testingtools import ON_CI_CLOUD




def setup_module():
    setup_catalogue('test_peeler', dataset_name='olfactory_bulb')
    setup_catalogue('test_peeler2', dataset_name='olfactory_bulb', duration=16.)

def teardown_module():
    shutil.rmtree('test_peeler')
    shutil.rmtree('test_peeler2')


def open_catalogue_window():
    dataio = DataIO(dirname='test_peeler')
    catalogueconstructor = CatalogueConstructor(dataio=dataio)
    app = pg.mkQApp()
    win = CatalogueWindow(catalogueconstructor)
    win.show()
    app.exec_()


def test_peeler_geometry():
    dataio = DataIO(dirname='test_peeler')
    
    catalogue0 = dataio.load_catalogue(chan_grp=0)
    # catalogue1 = dataio.load_catalogue(chan_grp=0, name='with_oversampling')
    
        
        
    # for catalogue in (catalogue0, catalogue1):
    for catalogue in (catalogue0, ):
        print()
        print('engine=geometrical')
        print('inter_sample_oversampling', catalogue['inter_sample_oversampling'])
        peeler = Peeler(dataio)

        peeler.change_params(engine='geometrical',
                                    catalogue=catalogue,
                                    chunksize=1024)
                                    
                                    

        t1 = time.perf_counter()
        peeler.run(progressbar=False)
        t2 = time.perf_counter()
        print('peeler.run_loop', t2-t1)

        spikes = dataio.get_spikes(chan_grp=0).copy()
        labels = catalogue['clusters']['cluster_label']
        count_by_label = [np.sum(spikes['cluster_label'] == label) for label in labels]
        print(labels)
        print(count_by_label)


@pytest.mark.skipif(ON_CI_CLOUD, reason='ON_CI_CLOUD')
def test_peeler_geometry_cl():
    dataio = DataIO(dirname='test_peeler')
    #~ catalogue = dataio.load_catalogue(chan_grp=0)
    catalogue0 = dataio.load_catalogue(chan_grp=0)
    # catalogue1 = dataio.load_catalogue(chan_grp=0, name='with_oversampling')
    
    # for catalogue in (catalogue0, catalogue1):
    for catalogue in (catalogue0, ):
        print()
        print('engine=geometrical_opencl')
        print('inter_sample_oversampling', catalogue['inter_sample_oversampling'])

        peeler = Peeler(dataio)
        
        catalogue['clean_peaks_params']['alien_value_threshold'] = None
        
        
        peeler.change_params(engine='geometrical_opencl',
                                    catalogue=catalogue,
                                    chunksize=1024,
                                    speed_test_mode=True,
                                    )

        t1 = time.perf_counter()
        peeler.run(progressbar=False)
        t2 = time.perf_counter()
        print('peeler.run_loop', t2-t1)

        spikes = dataio.get_spikes(chan_grp=0).copy()
        labels = catalogue['clusters']['cluster_label']
        count_by_label = [np.sum(spikes['cluster_label'] == label) for label in labels]
        print(labels)
        print(count_by_label)
        
        
        run_times = peeler.get_run_times(chan_grp=0, seg_num=0)
        print(run_times)


    
@pytest.mark.skipif(ON_CI_CLOUD, reason='TOO_OLD')
def test_peeler_empty_catalogue():
    """
    This test peeler with empty catalogue.
    This is like a peak detector.
    Check several chunksize and compare to offline-one-buffer.
    
    THIS TEST IS TOO OLD need to be rewritten
    
    """
    dataio = DataIO(dirname='test_peeler')
    #~ print(dataio)
    catalogue = dataio.load_catalogue(chan_grp=0)
    
    # empty catalogue for debug peak detection
    s = catalogue['centers0'].shape
    empty_centers = np.zeros((0, s[1], s[2]), dtype='float32')
    catalogue['centers0'] = empty_centers
    catalogue['centers1'] = empty_centers
    catalogue['centers2'] = empty_centers
    catalogue['cluster_labels'] = np.zeros(0, dtype=catalogue['cluster_labels'].dtype)
        
    
    sig_length = dataio.get_segment_length(0)
    chunksizes = [ 101, 174, 512, 1024, 1023, 10000, 150000]
    #~ chunksizes = [1024,]
    
    previous_peak = None
    
    for chunksize in chunksizes:
        print('**',  chunksize, '**')
        peeler = Peeler(dataio)
        peeler.change_params(engine='geometrical', catalogue=catalogue,chunksize=chunksize, 
                save_bad_label=True)
        t1 = time.perf_counter()
        #~ peeler.run(progressbar=False)
        #~ peeler.run_offline_loop_one_segment(seg_num=0, progressbar=False)
        peeler.run(progressbar=False)
        t2 = time.perf_counter()
        
        #~ print('n_side', peeler.n_side, 'n_span', peeler.n_span, 'peak_width', peeler.peak_width)
        #~ print('peeler.run_loop', t2-t1)
        
        spikes = dataio.get_spikes(seg_num=0, chan_grp=0)
        labeled_spike = spikes[spikes['cluster_label']>=0]
        unlabeled_spike = spikes[spikes['cluster_label']<0]
        assert labeled_spike.size == 0
        print(unlabeled_spike.size)
        
        is_sorted = np.all(np.diff(unlabeled_spike['index'])>=0)
        
        
        
        online_peaks = unlabeled_spike['index']
        engine = peeler.peeler_engine
        
        i_stop = sig_length-catalogue['signal_preprocessor_params']['pad_width']-engine.extra_size+engine.n_span
        sigs = dataio.get_signals_chunk(signal_type='processed', i_stop=i_stop)
        offline_peaks  = detect_peaks_in_chunk(sigs, engine.n_span, engine.relative_threshold, engine.peak_sign)
        print(offline_peaks.size)
        
        offline_peaks  = offline_peaks[offline_peaks<=online_peaks[-1]]
        
        assert offline_peaks.size == online_peaks.size
        np.testing.assert_array_equal(offline_peaks, online_peaks)
        
        if previous_peak is not None:
            last = min(previous_peak[-1], online_peaks[-1])
            previous_peak = previous_peak[previous_peak<=last]
            online_peaks_cliped = online_peaks[online_peaks<=last]
            assert  previous_peak.size == online_peaks_cliped.size
            np.testing.assert_array_equal(previous_peak, online_peaks_cliped)
        
        previous_peak = online_peaks
        


@pytest.mark.skipif(ON_CI_CLOUD, reason='To hard for CI')
def test_peeler_several_chunksize():
    
    dataio = DataIO(dirname='test_peeler')
    print(dataio)
    catalogue = dataio.load_catalogue(chan_grp=0)
    
    
    all_spikes = []
    sig_length = dataio.get_segment_length(0)
    chunksizes = [ 174, 512, 1024, 1023, 10000, 150000]
    #~ chunksizes = [512, 1024,]
    for chunksize in chunksizes:
        print('**',  chunksize, '**')
        peeler = Peeler(dataio)
        peeler.change_params(engine='geometrical', catalogue=catalogue,chunksize=chunksize)
        t1 = time.perf_counter()
        peeler.run(progressbar=False)
        t2 = time.perf_counter()
        print('extra_size', peeler.peeler_engine.extra_size, 'n_span', peeler.peeler_engine.n_span,
                        'peak_width', peeler.peeler_engine.peak_width)
        print('peeler.run_loop', t2-t1)
        
        # copy is need because the memmap is reset at each loop
        spikes = dataio.get_spikes(seg_num=0, chan_grp=0).copy()
        all_spikes.append(spikes)
        print(spikes.size)
    
    
    # clip to last spike
    last = min([spikes[-1]['index'] for spikes in all_spikes])
    for i, chunksize in enumerate(chunksizes):
        spikes = all_spikes[i]
        all_spikes[i] = spikes[spikes['index']<=last]
    
    previsous_spikes = None
    for i, chunksize in enumerate(chunksizes):
        print('**',  chunksize, '**')
        spikes = all_spikes[i]
        is_sorted = np.all(np.diff(spikes['index'])>=0)
        assert is_sorted
        
        labeled_spike = spikes[spikes['cluster_label']>=0]
        unlabeled_spike = spikes[spikes['cluster_label']<0]
        print('labeled_spike.size', labeled_spike.size, 'unlabeled_spike.size', unlabeled_spike.size)
        print(spikes)
        
        # TODO: Peeler chunksize influence the number of spikes
        
        if previsous_spikes is not None:
            assert  previsous_spikes.size == spikes.size
            np.testing.assert_array_equal(previsous_spikes['index'], spikes['index'])
            np.testing.assert_array_equal(previsous_spikes['cluster_label'], spikes['cluster_label'])
        
        previsous_spikes = spikes



def test_peeler_with_and_without_preprocessor():
    
    if ON_CI_CLOUD:
        engines = ['geometrical']
    else:
        engines = ['geometrical', 'geometrical_opencl']

    
    #~ engines = ['geometrical_opencl']
    
    for engine in engines:
        for i in range(2):
        #~ for i in [1]:
        
            print()
            if i == 0:
                print(engine, 'without processing')
                dataio = DataIO(dirname='test_peeler')
            else:
                print(engine, 'with processing')
                dataio = DataIO(dirname='test_peeler2')
            
            catalogue = dataio.load_catalogue(chan_grp=0)
            
            peeler = Peeler(dataio)
            peeler.change_params(engine=engine, catalogue=catalogue, chunksize=1024)
            t1 = time.perf_counter()
            peeler.run(progressbar=False)
            t2 = time.perf_counter()
            print('peeler run_time', t2 - t1)
            spikes = dataio.get_spikes(chan_grp=0).copy()
            labels = catalogue['clusters']['cluster_label']
            count_by_label = [np.sum(spikes['cluster_label'] == label) for label in labels]
            print(labels)
            print(count_by_label)
            
    
    


def open_PeelerWindow():
    dataio = DataIO(dirname='test_peeler')
    #~ dataio = DataIO(dirname='test_peeler2')
    initial_catalogue = dataio.load_catalogue(chan_grp=0)

    app = pg.mkQApp()
    win = PeelerWindow(dataio=dataio, catalogue=initial_catalogue)
    win.show()
    app.exec_()


def test_export_spikes():
    dataio = DataIO(dirname='test_peeler')
    dataio.export_spikes()







def debug_compare_peeler_engines():
    # this do not work because oversampling is not handle
    dataio = DataIO(dirname='test_peeler')
    print(dataio)
    
    
    engine_list = [ 
            ('geometrical argmin opencl', 'geometrical', {}),
            ('geometrical_opencl', 'geometrical_opencl', {}),
        ]
    
    all_spikes =  []
    for name, engine, kargs in engine_list:
        #~ print()
        #~ print(name)
        # catalogue = dataio.load_catalogue(chan_grp=0, name='with_oversampling')
        catalogue = dataio.load_catalogue(chan_grp=0)


        peeler = Peeler(dataio)
        peeler.change_params(engine=engine, catalogue=catalogue,chunksize=1024, **kargs)

        t1 = time.perf_counter()
        peeler.run(progressbar=False, duration=None)
        t2 = time.perf_counter()
        print(name, 'run', t2-t1)
        
        
        spikes = dataio.get_spikes(chan_grp=0).copy()
        #~ print(spikes.size)
        all_spikes.append(spikes)
        
        
        #~ print(dataio.get_spikes(chan_grp=0).size)
    
    print()
    #~ all_spikes[0] = all_spikes[0][88+80:88+81+10]
    #~ all_spikes[1] = all_spikes[1][88+80:88+81+10]

    #~ all_spikes[0] = all_spikes[0][:88+81]
    #~ all_spikes[1] = all_spikes[1][:88+81]
    
    labels = catalogue['clusters']['cluster_label']
    
    for i, spikes in enumerate(all_spikes):
        name = engine_list[i][0]
        print()
        print(name)
        print(spikes[:10])
        print(spikes.size)
        
        count_by_label = [np.sum(spikes['cluster_label'] == label) for label in labels]
        print(count_by_label)
            
        
        #~ assert all_spikes[0].size == spikes.size
        #~ assert np.all(all_spikes[0]['index'] == spikes['index'])
        #~ assert np.all(all_spikes[0]['cluster_label'] == spikes['cluster_label'])
        #~ assert np.all(np.abs(all_spikes[0]['jitter'] - spikes['jitter'])<0.0001)
    

if __name__ =='__main__':
    #~ setup_module()
    
    #~ open_catalogue_window()
    
    
    #~ test_peeler_geometry()
    
    #~ test_peeler_geometry_cl()
    
    
    #~ test_peeler_empty_catalogue()
    
    test_peeler_several_chunksize()
    
    #~ test_peeler_with_and_without_preprocessor()
    
    #~ test_export_spikes()
    
    
    #~ debug_compare_peeler_engines()
    
    #~ open_PeelerWindow()
    
    #~ teardown_module()



from tridesclous import download_dataset, get_dataset


def test_download_dataset():
    
    download_dataset(name='locust')
    download_dataset(name='olfactory_bulb')

def test_get_dataset():
    data, sample_rate = get_dataset(name='locust')
    assert data.shape == (431548, 4)
    assert sample_rate == 15000.0
    
    data, sample_rate = get_dataset(name='olfactory_bulb')
    assert data.shape == (150000, 14)
    assert sample_rate == 10000.0
    
    
    data, sample_rate = get_dataset(name='purkinje')
    
    data, sample_rate = get_dataset(name='striatum_rat')
    



if __name__ == '__main__':
    test_download_dataset()
    test_get_dataset()


import os
import time
import configparser
from pathlib import Path

import numpy as np

from .dataio import DataIO
from . catalogueconstructor import CatalogueConstructor, _dtype_peak
from .autoparams import get_auto_params_for_catalogue

try:
    import h5py
    HAVE_H5PY = True
except:
    HAVE_H5PY = False
    
try:
    import spikeextractors as se
    import spiketoolkit as st
    HAVE_SPIKEINTERFACE = True
except:
    HAVE_SPIKEINTERFACE = True



def import_from_spykingcircus(data_filename, spykingcircus_dirname, tdc_dirname):
    
    
    
    assert HAVE_H5PY, 'h5py is not installed'
    assert not os.path.exists(tdc_dirname), 'tdc already exisst please remvoe {}'.format(tdc_dirname)
    
    if spykingcircus_dirname.endswith('/'):
        spykingcircus_dirname = spykingcircus_dirname[:-1]
    

    #parse spyking circus params file
    params_filename = spykingcircus_dirname + '.params'
    config = configparser.ConfigParser(inline_comment_prefixes='#')
    config.read(params_filename)
    
    #Set DataIO and source
    _supported_formats = ('raw_binary', 'mcs_raw_binary')
    file_fmormat = config['data']['file_format']
    assert file_fmormat  in _supported_formats, 'only  {} are supported'.format(_supported_formats)

    dtype = config['data']['data_dtype']
    nb_channels = config.getint('data', 'nb_channels')
    if file_fmormat == 'mcs_raw_binary':
        import re
        data_filename = 'patch1.raw'
        with open(data_filename, 'rb') as f:
            header = f.read(5000).decode('Windows-1252')
        data_offset = re.search("EOH\r\n", header).start() + 5
    else:
        data_offset = config.getint('data', 'data_offset')
    sample_rate = config.getfloat('data', 'sampling_rate')
    
    dataio = DataIO(dirname=tdc_dirname)
    filenames = [data_filename]
    dataio.set_data_source(type='RawData', filenames=filenames, dtype=dtype,
                                     total_channel=nb_channels, sample_rate=sample_rate, offset=data_offset)
    
    #set probe file (supposed to be in same path)
    probe = config.get('data', 'mapping')
    probe_filename = os.path.join(os.path.dirname(data_filename), probe)
    dataio.set_probe_file(probe_filename)
    
    
    cc = CatalogueConstructor(dataio=dataio)
    
    # filter
    if config.getboolean('filtering', 'filter'):
        highpass_freq = config.getfloat('filtering', 'cut_off')
    else:
        highpass_freq = None
    common_ref_removal = config.getboolean('filtering', 'remove_median')
    
    # detection
    relative_threshold = config.getfloat('detection', 'spike_thresh')
    if config.get('detection', 'peaks') == 'negative':
        peak_sign='-'
    elif config.get('detection', 'peaks') == 'positive':
        peak_sign='+'
    else:
        raise(NotImlementedError)
    
    
    cc.set_global_params(
            chunksize=1024,
            memory_mode='memmap',
            mode='sparse',
            adjacency_radius_um=400,
            sparse_threshold=1.5)
    
    # params preprocessor
    cc.set_preprocessor_params(
            engine='numpy',
            highpass_freq=highpass_freq, 
            lowpass_freq=None,
            common_ref_removal=common_ref_removal,
            lostfront_chunksize=-1,
        )
    
    # params peak detector
    cc.set_peak_detector_params(
            method='global',
            engine='numpy',
            peak_sign=peak_sign, 
            relative_threshold=relative_threshold,
            peak_span=1./sample_rate,
            )

    
    t1 = time.perf_counter()
    duration=30.
    cc.estimate_signals_noise(seg_num=0, duration=duration)
    t2 = time.perf_counter()
    print('estimate_signals_noise', t2-t1)
    
    duration = dataio.get_segment_length(0)/dataio.sample_rate
    t1 = time.perf_counter()
    cc.run_signalprocessor(duration=duration, detect_peak=False)
    t2 = time.perf_counter()
    print('run_signalprocessor', t2-t1)
    
    
    
    #read peaks from results files
    name = os.path.basename(spykingcircus_dirname)
    result_filename = os.path.join(spykingcircus_dirname, '{}.result.hdf5'.format(name))
    result = h5py.File(result_filename,'r')
    
    #~ result.visit(lambda p: print(p))
    
    all_peaks = []
    for k in result['spiketimes'].keys():
        #~ print('k', k)
        _, label = k.split('_')
        label = int(label)
        indexes = np.array(result['spiketimes'][k])
        peaks = np.zeros(indexes.shape, dtype=_dtype_peak)
        peaks['index'][:] = indexes
        peaks['cluster_label'][:] = label
        peaks['segment'][:] = 0
        
        all_peaks.append(peaks)
        
    
    all_peaks = np.concatenate(all_peaks)
    order = np.argsort(all_peaks['index'])
    all_peaks = all_peaks[order]
    
    nb_peak = all_peaks.size
    
    cc.arrays.create_array('all_peaks', _dtype_peak, (nb_peak,), 'memmap')
    cc.all_peaks[:] = all_peaks
    cc.on_new_cluster()
    
    
    N_t = config.getfloat('data', 'N_t')
    n_right = int(N_t*sample_rate/1000.)//2
    n_left = -n_right
    
    cc.set_waveform_extractor_params(n_left=n_left, n_right=n_right)
    cc.sample_some_peaks( mode='rand', nb_max=10000)

    cc.extract_some_noise()
    cc.extract_some_features(method='peak_max')
    
    self.all_peaks['cluster_label'][cc.some_peaks_index] = all_peaks[cc.some_peaks_index]['cluster_label']
    cc.on_new_cluster()
    cc.compute_all_centroid()
    cc.refresh_colors()
    
    return cc
    

def import_from_spike_interface(recording, sorting, tdc_dirname, spike_per_cluster=300, align_peak=True, catalogue_params={}):
    output_folder = Path(tdc_dirname)
    
    # save prb file:
    probe_file = output_folder / 'probe.prb'
    se.save_to_probe_file(recording, probe_file)

    # save binary file (chunk by hcunk) into a new file
    raw_filename = output_folder / 'raw_signals.raw'
    n_chan = recording.get_num_channels()
    chunksize = 2**24// n_chan
    se.write_to_binary_dat_format(recording, raw_filename, time_axis=0, dtype='float32', chunk_size=chunksize)
    dtype='float32'
    offset = 0
    sr = recording.get_sampling_frequency()

    # initialize source and probe file
    dataio = DataIO(dirname=str(output_folder))
    nb_chan = recording.get_num_channels()

    dataio.set_data_source(type='RawData', filenames=[str(raw_filename)],
                               dtype=dtype, sample_rate=sr,
                               total_channel=nb_chan, offset=offset)
    dataio.set_probe_file(str(probe_file))
    print(dataio)
    
    params = get_auto_params_for_catalogue(dataio)
    for k in params:
        if k in catalogue_params:
            if isinstance(catalogue_params[k], dict):
                params[k].update(catalogue_params[k])
            else:
                params[k] = catalogue_params[k]
    
    cc = CatalogueConstructor(dataio=dataio)

    # global params
    d = {k:params[k] for k in ('chunksize', 'mode', )}
    cc.set_global_params(**d)

    # params preprocessor
    d = dict(params['preprocessor'])
    cc.set_preprocessor_params(**d)
    
    # params peak detector
    d = dict(params['peak_detector'])
    cc.set_peak_detector_params(**d)
    
    t1 = time.perf_counter()
    noise_duration = min(10., params['duration'], dataio.get_segment_length(seg_num=0)/dataio.sample_rate*.99)
    t1 = time.perf_counter()
    cc.estimate_signals_noise(seg_num=0, duration=noise_duration)
    t2 = time.perf_counter()
    print('estimate_signals_noise', t2-t1)
    
    duration = dataio.get_segment_length(0)/dataio.sample_rate
    t1 = time.perf_counter()
    cc.run_signalprocessor(duration=duration, detect_peak=False)
    t2 = time.perf_counter()
    print('run_signalprocessor', t2-t1)
    
    wf_left_ms = params['extract_waveforms']['wf_left_ms']
    wf_right_ms = params['extract_waveforms']['wf_right_ms']
    n_left = int(wf_left_ms / 1000. * dataio.sample_rate)
    n_right = int(wf_right_ms / 1000. * dataio.sample_rate)
    
    sig_size = dataio.get_segment_length(seg_num=0)
    
    peak_shift = {}
    extremum_channel = {}

    t1 = time.perf_counter()
    # compute the shift to have the true max at -n_left
    for label in sorting.get_unit_ids():
        indexes = sorting.get_unit_spike_train(label)
        indexes = indexes[indexes<(sig_size-n_right-1)]
        indexes = indexes[indexes>(-n_left+1)]
        if indexes.size>spike_per_cluster:
            keep = np.random.choice(indexes.size, spike_per_cluster, replace=False)
            indexes = indexes[keep]
        wfs = dataio.get_some_waveforms(seg_num=0, chan_grp=0, peak_sample_indexes=indexes, n_left=n_left, n_right=n_right)
        wf0 = np.median(wfs, axis=0)
        # maybe use '-' or '+'
        chan_max = np.argmax(np.max(np.abs(wf0), axis=0))
        ind_max = np.argmax(np.abs(wf0[:, chan_max]))
        extremum_channel[label] = chan_max
        shift = ind_max + n_left
        peak_shift[label] = shift
        #~ print(label,'>', shift)
    t2 = time.perf_counter()
    print('template', t2-t1)
    
    
    all_peaks = []
    for label in sorting.get_unit_ids():
        indexes = sorting.get_unit_spike_train(label)
        indexes = indexes[indexes<(sig_size-n_right-1)]
        indexes = indexes[indexes>(-n_left+1)]
        peaks = np.zeros(indexes.shape, dtype=_dtype_peak)
        if align_peak:
            indexes = indexes + peak_shift[label]
        peaks['index'][:] = indexes
        peaks['cluster_label'][:] = label
        peaks['segment'][:] = 0
        peaks['channel'][:] = extremum_channel[label]
        
        
        
        all_peaks.append(peaks)
    
    all_peaks = np.concatenate(all_peaks)
    order = np.argsort(all_peaks['index'])
    all_peaks = all_peaks[order]

    nb_peak = all_peaks.size
    
    cc.arrays.create_array('all_peaks', _dtype_peak, (nb_peak,), 'memmap')
    cc.all_peaks[:] = all_peaks
    cc.on_new_cluster()
    #~ print(cc.clusters)
    
    #  waveform per cluster selection
    t1 = time.perf_counter()
    global_keep = np.zeros(all_peaks.size, dtype='bool')
    for label in sorting.get_unit_ids():
        inds, = np.nonzero(all_peaks['cluster_label'] == label)
        if inds.size>spike_per_cluster:
            incluster_sel = np.random.choice(inds.size, spike_per_cluster, replace=False)
            inds = inds[incluster_sel]
        global_keep[inds] = True
    index, = np.nonzero(global_keep)

    cc.set_waveform_extractor_params(n_left=n_left, n_right=n_right)
    
    t1 = time.perf_counter()
    cc.sample_some_peaks(mode='force', index=index)
    t2 = time.perf_counter()
    print('sample_some_peaks', t2-t1)
    
    cc.clean_peaks(alien_value_threshold=None)
    
    cc.extract_some_noise()
    
    cc.extract_some_features(method=params['feature_method'], **params['feature_kargs'])
    
    # put back label 
    cc.all_peaks['cluster_label'][cc.some_peaks_index] = all_peaks[cc.some_peaks_index]['cluster_label']
    t1 = time.perf_counter()
    cc.on_new_cluster()
    cc.compute_all_centroid()
    t2 = time.perf_counter()
    print('compute_all_centroid', t2-t1)
    cc.refresh_colors()
    print(cc)
    
    return cc









from .myqt import QT
import pyqtgraph as pg

import os
from collections import OrderedDict

import tempfile

from ..dataio import DataIO
from ..datasource import data_source_classes
from .tools import get_dict_from_group_param
from ..tools import download_probe, fix_prb_file_py2
from ..probe_list import probe_list
from .probegeometryview import ProbeGeometryView




class InitializeDatasetWindow(QT.QDialog):
    def __init__(self, parent=None):
        QT.QDialog.__init__(self, parent = parent)
        
        self.setWindowTitle('Initailize new tridesclous')
        self.setModal(True)
        
        self.resize(600, 300)

        layout = QT.QVBoxLayout()
        self.setLayout(layout)
        
        #for step 'source_type'
        self.tree_params = pg.parametertree.ParameterTree(parent=self)
        self.tree_params.header().hide()
        layout.addWidget(self.tree_params)
        self.tree_params.hide()
        
        # for filenames
        self.but_addfiles = QT.QPushButton('Add files')
        layout.addWidget(self.but_addfiles)
        self.but_addfiles.clicked.connect(self.on_addfiles)
        self.but_addfiles.hide()
        self.list_files = QT.QListWidget()
        layout.addWidget(self.list_files)
        self.list_files.hide()
        
        # for chan_grp
        self.changroup_widget = ChannelGroupWidget()
        layout.addWidget(self.changroup_widget)
        self.changroup_widget.hide()
        
        # for dirname
        self.but_dirname = QT.QPushButton('change path')
        layout.addWidget(self.but_dirname)
        self.but_dirname.clicked.connect(self.on_change_path)
        self.but_dirname.hide()
        self.label_path = QT.QLabel()
        layout.addWidget(self.label_path)
        self.label_path.hide()
        self.edit_dirname = QT.QLineEdit()
        layout.addWidget(self.edit_dirname)
        self.edit_dirname.hide()
        
        layout.addStretch()
        
        self.but_next = QT.QPushButton('Next >')
        layout.addWidget(self.but_next)
        self.but_next.clicked.connect(self.validate_step)
        
        self.steps = ['source_type', 'filenames', 'source_params', 'mk_chan_grp', 'dirname']
        self.display_step(self.steps[0])
        
        self.final_params = {}
    
    
    def display_step(self, step):
        self.actual_step = step
        
        if step=='source_type':
            source_types = list(data_source_classes.keys())
            source_types.remove('InMemory')
            #~ if 'RawBinarySignal' in source_types:
                #~ source_types.remove('RawData')
            
            params = [{'name': 'source_type', 'type': 'list', 'values':source_types},
                            ]
            self.step_params = pg.parametertree.Parameter.create(name='Select source type', type='group', children = params)
            self.tree_params.setParameters(self.step_params, showTop=True)
            self.tree_params.show()
        
        elif step=='filenames':
            self.but_addfiles.show()
            self.list_files.show()
        
        elif step=='source_params':
            params = self.datasource_class.gui_params
            self.step_params = pg.parametertree.Parameter.create(name='Set params', type='group', children=params)
            self.tree_params.setParameters(self.step_params, showTop=True)
            self.tree_params.show()

        elif step=='mk_chan_grp':
            source_params = self._get_source_param()
            
            #~ try:
            if True:
                channel_names = self.datasource_class(**source_params).get_channel_names()
            #~ except:
                #~ #TYODO warning
                #~ print('Error file')
                #~ return
            print(channel_names)
            self.changroup_widget.set_channel_names(channel_names)
            self.changroup_widget.show()
        
        elif step=='dirname':
            self.but_dirname.show()
            self.label_path.show()
            self.edit_dirname.show()
            
            names = self.final_params['file_or_dir_names']
            name0 = names[0]
            suggested_dir = os.path.dirname(name0)
            if len(names)==1:
                name, _ = os.path.splitext(os.path.basename(name0))
                suggested_name = 'tdc_'+name
            else:
                suggested_name = 'tdc_'
            
            self.label_path.setText(suggested_dir)
            self.edit_dirname.setText(suggested_name)
        
    def _get_source_param(self):
        source_params = dict(self.final_params['source_params'])
        if self.datasource_class.mode =='one-file':
            source_params['filename'] = self.final_params['file_or_dir_names'][0]
        elif self.datasource_class.mode =='multi-file':
            source_params['filenames'] = self.final_params['file_or_dir_names']
        elif self.datasource_class.mode == 'one-dir':
            source_params['dirname'] = self.final_params['file_or_dir_names'][0]
        elif self.datasource_class.mode == 'multi-dir':
            source_params['dirnames'] = self.final_params['file_or_dir_names']
        print('source_params', source_params)
        return source_params
   
    def validate_step(self):
        #~ print('validate_step', self.actual_step)
        step = self.actual_step
        if step=='source_type':
            self.final_params['source_type'] = self.step_params['source_type']
            self.tree_params.hide()
            self.display_step('filenames')
            self.datasource_class = data_source_classes[self.final_params['source_type']]
            
        elif step=='filenames':
            file_or_dir_names = [self.list_files.item(i).text() for i in range(self.list_files.count())]
            if len(file_or_dir_names)==0:
                return
            if self.datasource_class.mode == 'one-file' and len(file_or_dir_names)!=1:
                return
            if self.datasource_class.mode == 'one-dir' and len(file_or_dir_names)!=1:
                return
            
            self.final_params['file_or_dir_names'] = file_or_dir_names
            self.but_addfiles.hide()
            self.list_files.hide()
            
            if self.datasource_class.gui_params is None:
                self.final_params['source_params'] = {}
                self.display_step('mk_chan_grp')
            else:
                self.display_step('source_params')
        
        elif step=='source_params':
            self.final_params['source_params'] = get_dict_from_group_param(self.step_params)
            self.tree_params.hide()
            self.display_step('mk_chan_grp')
        
        elif step=='mk_chan_grp':
            #~ print(self.final_params)
            self.final_params['channel_groups'] = self.changroup_widget.channel_groups
            self.changroup_widget.hide()
            self.display_step('dirname')
        
        elif step=='dirname':
            p1 = self.label_path.text()
            p2 = self.edit_dirname.text()
            dirname = os.path.join(p1, p2)
            if DataIO.check_initialized(dirname):
                #todo warning
                print('already exist')
                return
            self.final_params['dirname'] = dirname
            self.but_dirname.hide()
            self.label_path.hide()
            self.edit_dirname.hide()
            self.make_it()
            self.accept()
    
    def on_addfiles(self):
        print('on_addfiles')
        print(self.datasource_class)
        print(self.datasource_class.mode)
        
        if self.datasource_class.mode.endswith('-file'):
            fd = QT.QFileDialog(fileMode=QT.QFileDialog.ExistingFiles, acceptMode=QT.QFileDialog.AcceptOpen)
            fd.setNameFilters(['All (*)'])
        elif self.datasource_class.mode.endswith('-dir'):
            fd = QT.QFileDialog(fileMode=QT.QFileDialog.DirectoryOnly, acceptMode=QT.QFileDialog.AcceptOpen)
            #~ fd.setNameFilters(['All (*)'])
        
        fd.setViewMode(QT.QFileDialog.Detail)
        if fd.exec_():
            filenames = fd.selectedFiles()
            self.list_files.addItems(filenames)
    
    def on_change_path(self):
        dirname = self.label_path.text()
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.DirectoryOnly, acceptMode=QT.QFileDialog.AcceptOpen)
        fd.setViewMode(QT.QFileDialog.Detail)
        fd.setDirectory(os.path.dirname(dirname))
        if fd.exec_():
            dirname = fd.selectedFiles()[0]
            self.label_path.setText(dirname)
            #~ self.list_files.addItems(filenames)
            
    def make_it(self):
        try:
            p = self.final_params
            
            source_params = self._get_source_param()
            
            dataio = DataIO(p['dirname'])
            dataio.set_data_source(type=p['source_type'], **source_params)
            dataio.set_channel_groups(p['channel_groups'])
            print(dataio)
            self.dirname_created = p['dirname']
        except Exception as e:
            print(e)



class ChannelGroupWidget(QT.QWidget):
    def __init__(self, parent=None):
        QT.QWidget.__init__(self, parent = parent)
        
        layout = QT.QHBoxLayout()
        self.setLayout(layout)
        
        self.list_channel = QT.QListWidget(selectionMode=QT.QAbstractItemView.ExtendedSelection, selectionBehavior=QT.QTreeView.SelectRows)
        layout.addWidget(self.list_channel)

        v = QT.QVBoxLayout()
        
        but = QT.QPushButton('clear')
        v.addWidget(but)
        but.clicked.connect(self.clear_table)
        
        layout.addLayout(v)
        but = QT.QPushButton('Set manual channel group from selection')
        v.addWidget(but)
        but.clicked.connect(self.add_chan_grp)

        but = QT.QPushButton('Set channel group from PRB file')
        v.addWidget(but)
        but.clicked.connect(self.open_prb_file)
        
        h = QT.QHBoxLayout()
        v.addLayout(h)
        but = QT.QPushButton('Download PRB file')
        but.clicked.connect(self.download_prb_file)
        h.addWidget(but)
        self.comboPrb = QT.QComboBox()
        h.addWidget(self.comboPrb)
        all_prb = ['']
        for origin, probe_names in probe_list.items():
            for probe_name in probe_names:
                all_prb.append('{}: {}'.format(origin, probe_name))
        self.comboPrb.addItems(all_prb)
        
        
        self.table_grp = QT.QTableWidget()
        v.addWidget(self.table_grp)

        but = QT.QPushButton('Show probes geometry')
        v.addWidget(but)
        but.clicked.connect(self.show_prb_geometry)
        
        self.channel_names = None
        self.channel_groups = None
        
        self.geometryview = None
    
    def set_channel_names(self, channel_names):
        self.n = len(channel_names)
        self.channel_names = channel_names
        self.list_channel.clear()
        self.list_channel.addItems(channel_names)
        
        channels = list(range(self.n))
        #~ geometry = { c: [0, i] for i, c in enumerate(channels) }
        geometry = None
        self.channel_groups = OrderedDict()
        self.channel_groups[0] = dict(channels=channels, geometry=geometry)
        
        self.refresh_table()
        

    def refresh_table(self):
        t = self.table_grp
        
        t.clear()
        t.setColumnCount(3)
        t.setRowCount(len(self.channel_groups))
        
        t.setHorizontalHeaderLabels(['key', 'nb_chan', 'channels'])
        for i, k in enumerate(self.channel_groups.keys()):
            channel_group = self.channel_groups[k]
            channels = channel_group['channels']
            t.setItem(i, 0,  QT.QTableWidgetItem('{}'.format(k)))
            t.setItem(i, 1,  QT.QTableWidgetItem('{}'.format(len(channels))))
            txt = ' '.join([str(c) for c in channels])
            t.setItem(i, 2,  QT.QTableWidgetItem('[{}]'.format(txt)))
        
        t.resizeColumnsToContents()
            
    def add_chan_grp(self):
        channels = [ind.row() for ind in self.list_channel.selectedIndexes()]
        if len(channels)==0: return
        k = len(self.channel_groups)
        #~ geometry = { c: [0, i] for i, c in enumerate(channels) }
        geometry = None
        self.channel_groups[k] = dict(channels=channels, geometry=geometry)
        self.refresh_table()
    
    def open_prb_file(self):
        fd = QT.QFileDialog(fileMode=QT.QFileDialog.ExistingFiles, acceptMode=QT.QFileDialog.AcceptOpen)
        fd.setNameFilters(['prb (*.prb *.PRB)', 'All (*)'])
        fd.setViewMode(QT.QFileDialog.Detail)
        if fd.exec_():
            prb_filename = fd.selectedFiles()[0]
            
            fix_prb_file_py2(prb_filename)
            
            d = {}
            exec(open(prb_filename).read(), None, d)
            
            self.channel_groups = OrderedDict()
            self.channel_groups.update(d['channel_groups'])
            self.refresh_table()
    
    def download_prb_file(self):
        probe_name = str(self.comboPrb.currentText())
        if len(probe_name) == 0:
            return
        origin, probe_name = probe_name.split(': ')
        #~ print(origin, probe_name)
        local_dirname = tempfile.gettempdir()
        #~ print(local_dirname)
        prb_filename = download_probe(local_dirname, probe_name, origin=origin)
        #~ print(prb_filename)

        d = {}
        exec(open(prb_filename).read(), None, d)
        
        #~ print()
        self.channel_groups = OrderedDict()
        self.channel_groups.update(d['channel_groups'])
        self.refresh_table()

    
    def clear_table(self):
        self.channel_groups = OrderedDict()
        self.refresh_table()

    def show_prb_geometry(self):
        if self.geometryview is not None:
            self.geometryview.close()
        
        self.geometryview = ProbeGeometryView(channel_groups=self.channel_groups, parent=self)
        #~ self.geometryview.setModal(True)
        self.geometryview.setWindowFlags(QT.Qt.Window)
        
        self.geometryview.show()
        



from .myqt import QT
import pyqtgraph as pg

import numpy as np

from .. import labelcodes
from .base import WidgetBase




class ClusterBaseList(WidgetBase):
    """
    Base for ClusterPeakList (catalogue window) and ClusterSpikeList (Peeler window)
    """
    
    def __init__(self, controller=None, parent=None):
        WidgetBase.__init__(self, parent=parent, controller=controller)
        
        self.layout = QT.QVBoxLayout()
        self.setLayout(self.layout)
        
        h = QT.QHBoxLayout()
        self.layout.addLayout(h)
        h.addWidget(QT.QLabel('sort by'))
        self.combo_sort = QT.QComboBox()
        self.combo_sort.addItems(['label', 'extremum_channel', 'extremum_amplitude', 'waveform_rms', 'nb_peak'])
        self.combo_sort.currentIndexChanged.connect(self.refresh)
        h.addWidget(self.combo_sort)
        h.addStretch()
        
        self.table = QT.QTableWidget()
        self.layout.addWidget(self.table)
        self.table.itemChanged.connect(self.on_item_changed)
        self.table.cellDoubleClicked.connect(self.on_double_clicked)
        
        self.make_menu()
        
        self.refresh()

    def make_menu(self):
        raise(NotImplementedError)
    
    def refresh(self):
        self.table.itemChanged.disconnect(self.on_item_changed)
        
        self.table.clear()
        labels = ['cluster_label', 'show/hide', 'nb_peaks', 'extremum_channel', 'cell_label', 'tag', 'annotations']
        self.table.setColumnCount(len(labels))
        self.table.setHorizontalHeaderLabels(labels)
        #~ self.table.setMinimumWidth(100)
        #~ self.table.setColumnWidth(0,60)
        self.table.setContextMenuPolicy(QT.Qt.CustomContextMenu)
        self.table.customContextMenuRequested.connect(self.open_context_menu)
        self.table.setSelectionMode(QT.QAbstractItemView.ExtendedSelection)
        self.table.setSelectionBehavior(QT.QAbstractItemView.SelectRows)
        
        sort_mode = str(self.combo_sort.currentText())
        
        clusters = self.controller.clusters
        clusters = clusters[clusters['cluster_label']>=0]
        if sort_mode=='label':
            order =np.arange(clusters.size)
        elif sort_mode=='extremum_channel':
            order = np.argsort(clusters['extremum_channel'])
        elif sort_mode=='extremum_amplitude':
            order = np.argsort(np.abs(clusters['extremum_amplitude']))[::-1]
        elif sort_mode=='waveform_rms':
            order = np.argsort(clusters['waveform_rms'])[::-1]
        elif sort_mode=='nb_peak':
            order = np.argsort(clusters['nb_peak'])[::-1]
        
        cluster_labels = self._special_label + self.controller.positive_cluster_labels[order].tolist()
        
        self.table.setRowCount(len(cluster_labels))
        
        for i, k in enumerate(cluster_labels):
            color = self.controller.qcolors.get(k, QT.QColor( 'white'))
            pix = QT.QPixmap(16,16)
            pix.fill(color)
            icon = QT.QIcon(pix)
            
            if k<0:
                name = '{} ({})'.format(k, labelcodes.to_name[k])
            else:
                name = '{}'.format(k)
            item = QT.QTableWidgetItem(name)
            item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
            self.table.setItem(i,0, item)
            item.setIcon(icon)
            
            item = QT.QTableWidgetItem('')
            item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable|QT.Qt.ItemIsUserCheckable)
            
            item.setCheckState({ False: QT.Qt.Unchecked, True : QT.Qt.Checked}[self.controller.cluster_visible.get(k, False)])
            self.table.setItem(i,1, item)
            item.label = k

            item = QT.QTableWidgetItem('{}'.format(self.controller.cluster_count.get(k, 0)))
            item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
            self.table.setItem(i,2, item)
            
            c = self.controller.get_extremum_channel(k)
            if c is not None:
                item = QT.QTableWidgetItem('{}: {}'.format(c, self.controller.channel_names[c]))
                item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                self.table.setItem(i,3, item)
            
            if k>=0:
                clusters = self.controller.clusters
                ## ind = np.searchsorted(clusters['cluster_label'], k) ## wrong because searchsortedmust be ordered
                ind = np.nonzero(clusters['cluster_label'] == k)[0][0]
                
                for c, attr in enumerate(['cell_label', 'tag', 'annotations']):
                    value = clusters[attr][ind]
                    item = QT.QTableWidgetItem('{}'.format(value))
                    item.setFlags(QT.Qt.ItemIsEnabled|QT.Qt.ItemIsSelectable)
                    self.table.setItem(i,4+c, item)

            
        for i in range(5):
            self.table.resizeColumnToContents(i)
        self.table.itemChanged.connect(self.on_item_changed)        

    def on_item_changed(self, item):
        if item.column() != 1: return
        sel = {QT.Qt.Unchecked : False, QT.Qt.Checked : True}[item.checkState()]
        #~ k = self.controller.cluster_labels[item.row()]
        k = item.label
        self.controller.cluster_visible[k] = bool(item.checkState())
        self.cluster_visibility_changed.emit()
    
    def on_double_clicked(self, row, col):
        for k in self.controller.cluster_visible:
            self.controller.cluster_visible[k] = False
            
        k = self.table.item(row, 1).label
        self.controller.cluster_visible[k] = True
        self.refresh()
        self.cluster_visibility_changed.emit()
        
    
    def selected_cluster(self):
        selected = []
        #~ for index in self.table.selectedIndexes():
        for item in self.table.selectedItems():
            #~ if index.column() !=1: continue
            if item.column() != 1: continue
            #~ selected.append(self.controller.cluster_labels[index.row()])
            selected.append(item.label)
        return selected
    
    
    def open_context_menu(self):
        self.menu.popup(self.cursor().pos())
        #~ menu.exec_(self.cursor().pos())
    
    def show_all(self):
        for k in self.controller.cluster_visible:
            self.controller.cluster_visible[k] = True
        self.refresh()
        self.cluster_visibility_changed.emit()
    
    def hide_all(self):
        for k in self.controller.cluster_visible:
            self.controller.cluster_visible[k] = False
        self.refresh()
        self.cluster_visibility_changed.emit()
    
    

    


